#include "FreqMNAM.h"

FreqMNAM::FreqMNAM(const ExtDoubleVector& fvec, Circuit* cir,
const ElemFlag& mask)
{
  // resize freq_vec to the size of the fvec so
  // as to prevent segmentation faults
  freq_vec.resize(fvec.size());
  for (unsigned int i = 0; i < fvec.size(); i++)
    freq_vec[i] = fvec[i];
  current_findex = 0;

  // Set internal variables first.
  assert(cir);
  my_circuit = cir;
  mymask = mask;

  //----------------------------------------------------------------------
  // Fill the element vector.
  // Reserve a reasonable amount of memory.
  // This routine allocates some default memory to hold the elements
  // in the Element vector
  elem_vec.reserve(my_circuit->getNumberOfElements());

  //----------------------------------------------------------------------
  // Loop throw all selected elements in circuit.
  // the circuit pointer is pointing to the first element
  my_circuit->setFirstElement(mask);

  // ..and is now looping through all the elements in the circuit
  // if it finds new elements during looping, it adds it to the element vector
  Element* elem = my_circuit->nextElement();
  while(elem)
  {
    // Put the element in the vector
    elem_vec.push_back(elem);
    // get next element pointer
    elem = my_circuit->nextElement();
  }

  //----------------------------------------------------------------------
  // Local variables
  // Number of elements, the size of the element vector just created
  unsigned n_elem = elem_vec.size();

  // The number of frequencies, the size of the frequency vector
	// which is a constructor argument
  unsigned n_freqs = freq_vec.size();

  // Scratch variables
  unsigned findex;
  int error;

  //----------------------------------------------------------------------
  // Set the initial value for the MNAM dimension.
  dimension = 0;

  // Set row/column information on terminals
  // Also find the reference node(s)
  // Now it is time to expand the MNAM depending on the no of terminals
	// present. The circuit pointer goes through all the terminals and
	// simultaneously searches for the reference terminal. If a terminal
	// found is not a reference temrinal then the dimension of the matrix is
	// increased. If the terminal is a reference terminal, then the boolean
	// variable "found" is set to true.
  my_circuit->setFirstTerminal();
  Terminal* term = my_circuit->nextTerminal();
  bool found = false;
  while (term)
  {
    // check if the terminal is a reference
    if (term->isRef())
    {
      term->setRC(0);
      found = true;
    }
    else
    {
      // The dimension is the non-reference terminal count.
      dimension++;
      term->setRC(dimension);
    }
    // get next terminal
    term = my_circuit->nextTerminal();
  }
  assert(found);

  //----------------------------------------------------------------------
  // Get final dimension asking for extra rows and columns
  // the getExtraRC function is defined in Element.cc and it's function is to
  // accomodate any extra equations that the element may require. It finds out
  // this extra number and returns this number which is then added to our
  // dimension varable, so that the correct matrix size is created
  for (unsigned i=0; i < n_elem; i++)
    dimension += elem_vec[i]->getExtraRC(dimension + 1, FREQ_DOMAIN);

  //----------------------------------------------------------------------
  // Allocate space for the matrix vector.
  mnam_vec = new char*[n_freqs];

  //----------------------------------------------------------------------
  // Allocate memory for the vector of source vectors.
  // For this we need to know first the real dimension of the MNAM
  // Initialize the vectors to zero.
  // The source vector size depends on the number of frequencies specified
  // and is created accordingly. For each frequency there must be a 'new'
  // source vector and whose dimension will be equal to the dimension variable
  // calculated above.
  sourceV_vec = new double_complex*[n_freqs];
  for (findex = 0; findex < n_freqs; findex++)
    sourceV_vec[findex] = new double_complex[dimension];

  //----------------------------------------------------------------------
  // Create one matrix at each frequency and clear source vector
  // The source vector is initialized to zero at each frequency
  // Notice the use of nested for loops
  for (findex = 0; findex < n_freqs; findex++)
  {
    // Set source vector to zero
    for (unsigned i=0; i<dimension; i++)
      sourceV_vec[findex][i] = zero;

    //----------------------------------------------------------------------
    // This creates the MNAM vector using the SPARSE routine spCreate
    mnam_vec[findex] = spCreate(dimension, spCOMPLEX, &error);
    // For now check only using assertions. Later add exceptions
    assert(mnam_vec[findex]);
    assert(error == spOKAY);
  }

  //----------------------------------------------------------------------
  // Now fill the matrices
  // At each frequency, the MNAM is filled with the value in the Element vector
  for (current_findex = 0; current_findex < n_freqs; current_findex++)
  {
    for (unsigned i=0; i < n_elem; i++)
    {
      // Tell each element to fill the MNAM
      elem_vec[i]->fillMNAM(this);
    }
    // Use this to print the matrix for debugging
    // print();
  }
}


// Overloaded constructor with a DenseDoubleVector fvec
FreqMNAM::FreqMNAM(const DenseDoubleVector& fvec, Circuit* cir,
const ElemFlag& mask)
{
  // resize freq_vec to the size of the fvec so
  // as to prevent segmentation faults
  freq_vec.resize(fvec.size());
  mtl::copy(fvec, freq_vec);
  current_findex = 0;

  // Set internal variables first.
  assert(cir);
  my_circuit = cir;
  mymask = mask;

  //----------------------------------------------------------------------
  // Fill the element vector.
  // Reserve a reasonable amount of memory.
  // This routine allocates some default memory to hold the elements
  // in the Element vector
  elem_vec.reserve(my_circuit->getNumberOfElements());

  //----------------------------------------------------------------------
  // Loop throw all selected elements in circuit.
  // the circuit pointer is pointing to the first element
  my_circuit->setFirstElement(mask);

  // ..and is now looping through all the elements in the circuit
  // if it finds new elements during looping, it adds it to the element vector
  Element* elem = my_circuit->nextElement();
  while(elem)
  {
    // Put the element in the vector
    elem_vec.push_back(elem);
    // get next element pointer
    elem = my_circuit->nextElement();
  }

  //----------------------------------------------------------------------
  // Local variables
  // Number of elements, the size of the element vector just created
  unsigned n_elem = elem_vec.size();

  // The number of frequencies, the size of the frequency vector
  // which is a constructor argument
  unsigned n_freqs = freq_vec.size();

  // Scratch variables
  unsigned findex;
  int error;

  //----------------------------------------------------------------------
  // Set the initial value for the MNAM dimension.
  dimension = 0;

  // Set row/column information on terminals
  // Also find the reference node(s)
  // Now it is time to expand the MNAM depending on the no of terminals present
  // The circuit pointer goes through all the terminals and simultaneously searches
  // for the reference terminal. If an terminal found is not a reference temrinal
  // then the dimension of the matrix is increased. If the terminal is a reference
  // terminal, then the boolean variable "found" is set to true.
  my_circuit->setFirstTerminal();
  Terminal* term = my_circuit->nextTerminal();
  bool found = false;
  while (term)
  {
    // check if the terminal is a reference
    if (term->isRef())
    {
      term->setRC(0);
      found = true;
    }
    else
    {
      // The dimension is the non-reference terminal count.
      dimension++;
      term->setRC(dimension);
    }
    // get next terminal
    term = my_circuit->nextTerminal();
  }
  assert(found);

  //----------------------------------------------------------------------
  // Get final dimension asking for extra rows and columns
  // the getExtraRC function is defined in Element.cc and it's function is to
  // accomodate any extra equations that the element may require. It finds out
  // this extra number and returns this number which is then added to our
  // dimension varable, so that the correct matrix size is created
  for (unsigned i=0; i < n_elem; i++)
    dimension += elem_vec[i]->getExtraRC(dimension + 1, FREQ_DOMAIN);

  //----------------------------------------------------------------------
  // Allocate space for the matrix vector.
  mnam_vec = new char*[n_freqs];

  //----------------------------------------------------------------------
  // Allocate memory for the vector of source vectors.
  // For this we need to know first the real dimension of the MNAM
  // Initialize the vectors to zero.
  // The source vector size depends on the number of frequencies specified
  // and is created accordingly. For each frequency there must be a 'new'
  // source vector and whose dimension will be equal to the dimension variable
  // calculated above.
  sourceV_vec = new double_complex*[n_freqs];
  for (findex = 0; findex < n_freqs; findex++)
    sourceV_vec[findex] = new double_complex[dimension];

  //----------------------------------------------------------------------
  // Create one matrix at each frequency and clear source vector
  // The source vector is initialized to zero at each frequency
  // Notice the use of nested for loops
  for (findex = 0; findex < n_freqs; findex++)
  {
    // Set source vector to zero
    for (unsigned i=0; i<dimension; i++)
      sourceV_vec[findex][i] = zero;

    //----------------------------------------------------------------------
    // This creates the MNAM vector using the SPARSE routine spCreate
    mnam_vec[findex] = spCreate(dimension, spCOMPLEX, &error);
    // By now check only using assertions. Later add exceptions
    assert(mnam_vec[findex]);
    assert(error == spOKAY);
  }

  //----------------------------------------------------------------------
  // Now fill the matrices
  // At each frequency, the MNAM is filled with the value in the Element vector
  for (current_findex = 0; current_findex < n_freqs; current_findex++)
  {
    for (unsigned i=0; i < n_elem; i++)
    {
      // Tell each element to fill the MNAM
      elem_vec[i]->fillMNAM(this);
    }
    // Use this to print the matrix for debugging
    // print();
  }
}


FreqMNAM::~FreqMNAM()
{
  // Reverse the order of creation
  // Free matrices before matrix vector
  for (unsigned findex = 0; findex < freq_vec.size(); findex++)
    spDestroy(mnam_vec[findex]);

  for (unsigned findex = 0; findex < freq_vec.size(); findex++)
    delete [] sourceV_vec[findex];
  delete [] sourceV_vec;
  delete [] mnam_vec;
}


//--------------------------------------------------------------
// Methods related to the frequency vector
//--------------------------------------------------------------
void FreqMNAM::setFreqIndex(const unsigned& findex)
{
  assert(findex < freq_vec.size());
  current_findex = findex;
}

//----------------------------------------------------------------------
// The frequency vector is resetted here and the Source Vector and
// the MNAM is regenerated again. Basically, a new frequency vector
// is created and it is used to create the Source and MNAM vectors
// just like in the FreqMNAM constructor
void FreqMNAM::resetFreqVec(const ExtDoubleVector& fvec)
{
  // Assign the new vector. We support only same-size vectors
  assert(fvec.size() == freq_vec.size());
	for (unsigned int i = 0; i < fvec.size(); i++)
		freq_vec[i] = fvec[i];

  // Number of elements
  unsigned n_elem = elem_vec.size();
  // The number of frequencies
  unsigned n_freqs = freq_vec.size();

  // Now fill the matrices
  for (current_findex = 0; current_findex < n_freqs; current_findex++)
  {
    // Clean the matrix and source vector first
    for (unsigned i=0; i<dimension; i++)
      sourceV_vec[current_findex][i] = zero;
    spClear(mnam_vec[current_findex]);

    for (unsigned i=0; i < n_elem; i++)
		{
			// Tell each element to fill the MNAM
			elem_vec[i]->fillMNAM(this);
		}
  }
}

// Overloaded version with a DenseDoubleVector fvec
// argument
void FreqMNAM::resetFreqVec(const DenseDoubleVector& fvec)
{
  // Assign the new vector. We support only same-size vectors
  assert(fvec.size() == freq_vec.size());
  mtl::copy(fvec, freq_vec);

  // Number of elements
  unsigned n_elem = elem_vec.size();
  // The number of frequencies
  unsigned n_freqs = freq_vec.size();

  // Now fill the matrices
  for (current_findex = 0; current_findex < n_freqs; current_findex++)
  {
    // Clean the matrix and source vector first
    for (unsigned i=0; i<dimension; i++)
      sourceV_vec[current_findex][i] = zero;
    spClear(mnam_vec[current_findex]);

    for (unsigned i=0; i < n_elem; i++)
		{
			// Tell each element to fill the MNAM
			elem_vec[i]->fillMNAM(this);
		}
  }
}

//--------------------------------------------------------------
// Methods to operate the matrices
//--------------------------------------------------------------
unsigned FreqMNAM::getDim()
{
  return dimension;
}

//----------------------------------------------------------------------
// This re-orders and factors ALL the matrices if necessary, using SPARSE
// routines
void FreqMNAM::factorAll() throw(string&)
{
  // The number of frequencies
  int n_freqs = freq_vec.size();

  // Reorder and factor the matrices
  for (int findex = 0; findex < n_freqs; findex++)
  {
    factor(findex);
  }
}

//----------------------------------------------------------------------
// This re-orders and factors only one particular matrix depending on
// it's index number 'findex'. It also does some checking depending on
// values of some flags generated by SPARSE
void FreqMNAM::factor(const unsigned& findex) throw(string&)
{
  assert(findex < freq_vec.size());
  // Preorder the matrix taking advantage of it being an MNAM
  spMNA_Preorder(mnam_vec[findex]);

  // Factor
  spFactor(mnam_vec[findex]);
  // Check for errors
  int error = spError(mnam_vec[findex]);
  // Find what type of error
  if (error != spOKAY)
    switch (error)
    {
    case spSMALL_PIVOT:
		throw (string("MNAM is ill-conditioned."));

    case spZERO_DIAG:
		throw (string("MNAM has a zero-pivot."));

    case spSINGULAR:
		throw (string("MNAM is singular."));

    case spNO_MEMORY:
		throw (string("Out of memory after MNAM factorization."));

    default:
		throw (string("fatal error after MNAM factorization."));
    }
}

//----------------------------------------------------------------------
// It gets the source vector at a particular frequency
void FreqMNAM::getSource(const unsigned& findex, ExtComplexVector& sourceV)
{
  assert(findex < freq_vec.size());
	if (sourceV.size() != dimension)
		sourceV.resize(dimension);
  sourceV = ExtComplexVector(sourceV_vec[findex], dimension);
}

//----------------------------------------------------------------------
// It solves a matrix system for a fixed source vector, i.e. for a
// particular frequency. The solution is stored in the result vector.
// Here the size of the result vector is set to the value 'dimension'
void FreqMNAM::solve(const unsigned& findex, DenseComplexVector& result)
{
  assert(findex < freq_vec.size());
  // Check dimension of result vector
  if (result.size() != dimension)
    result.resize(dimension);
  // spSolve requires a double pointer
  // This trick should work with the normal complex type with (re, im)
  // members.
  spSolve(mnam_vec[findex], (double*)&(sourceV_vec[findex][0]),
	(double*)&(result[0]));
  assert(spError(mnam_vec[findex]) == spOKAY);
}

//----------------------------------------------------------------------
// It solves a matrix system for a given source vector
// The solution is stored in the result vector.
// Here the size of the result vector is set to the value 'dimension'
void FreqMNAM::solve(const unsigned& findex, const DenseComplexVector& sourceV,
DenseComplexVector& result)
{
	assert(findex < freq_vec.size());
  assert(sourceV.size() == dimension);
  // Check dimension of result vector
  if (result.size() != dimension)
    result.resize(dimension);
  // spSolve requires a double pointer
  // This trick should work with the normal complex type with (re, im)
  // members.
  spSolve(mnam_vec[findex], (double*)&(sourceV[0]), (double*)&(result[0]));
  assert(spError(mnam_vec[findex]) == spOKAY);
}

// overloaded version of above function with a different signature
void FreqMNAM::solve(const unsigned& findex, const ExtComplexVector& sourceV,
DenseComplexVector& result)
{
  assert(findex < freq_vec.size());
  assert(sourceV.size() == dimension);
  // Check dimension of result vector
  if (result.size() != dimension)
    result.resize(dimension);
  // spSolve requires a double pointer
  // This trick should work with the normal complex type with (re, im)
  // members.
  spSolve(mnam_vec[findex], (double*)&(sourceV[0]), (double*)&(result[0]));
  assert(spError(mnam_vec[findex]) == spOKAY);
}


//--------------------------------------------------------------
// Methods to fill the source vectors
//--------------------------------------------------------------
void FreqMNAM::addToSource(const unsigned& pos, const unsigned& neg,
const double_complex& val)
{
  assert(pos <= dimension);
  assert(neg <= dimension);
  assert(current_findex < freq_vec.size());
  if (pos)
    sourceV_vec[current_findex][pos-1] += val;
  if (neg)
    sourceV_vec[current_findex][neg-1] -= val;
}

//--------------------------------------------------------------
// Methods for debugging
//--------------------------------------------------------------
void FreqMNAM::print(int i)
{
  if (i == -1)
    i = current_findex;
  assert(unsigned(i) < freq_vec.size());
  cout << "Frequency: " << freq_vec[i] << endl;
  spPrint(mnam_vec[i], 0, 1, 1);
  cout << "Source Vector:\n[";
  for (unsigned j=0; j < dimension; j++)
    cout << sourceV_vec[i][j] << ", ";
  cout << "]\n\n";
}
