#include "SVTran2.h"
#include <fstream>
#include <iomanip>

extern "C"
{
#include "../inout/ftvec.h"
#include "../inout/report.h"
}

void buildTIncidence(ElemFlag mask, Circuit*& my_circuit,
                     IntDenseColMatrix& T, ElementVector& elem_vec,
                     int& n_states, int& max_n_states);

// Static members
const unsigned SVTran2::n_par = 10;

// Element information
ItemInfo SVTran2::ainfo =
{
  "SVTran2",
  "State-Variable-Based Time-Marching Transient Analysis",
  "Carlos E. Christoffersen",
  DEFAULT_ADDRESS
};

// Parameter information
ParmInfo SVTran2::pinfo[] =
{
  {"tstop", "Stop time (s)", TR_DOUBLE, false},
  {"tstep", "Time step (s)", TR_DOUBLE, false},
  {"nst", "No save time (s)", TR_DOUBLE, false},
  {"deriv", "Approximate derivatives or use automatic diff.", TR_INT, false},
  {"msv", "Use Msv flag", TR_BOOLEAN, false},
  {"im", "Integration method (1=trapezoidal) (2=backward euler)", TR_INT, false},
  {"permc_spec", "Permutation ordering to factor Msv (0, 1 or 2)",
   TR_INT, false},
  {"out_steps", "Number of steps skipped for output simulation progress",
   TR_INT, false},
  {"gcomp", "Compensation network conductance (S)", TR_DOUBLE, false},
  {"opt", "run a DC analysis up front", TR_BOOLEAN, false}
};


SVTran2::SVTran2() : Analysis(&ainfo, pinfo, n_par), ls_size(0),
    superLU(false)
{
  // Parameter stuff
  paramvalue[0] = &(tf);
  paramvalue[1] = &(h);
  paramvalue[2] = &(nst = zero);
  paramvalue[3] = &(deriv = 0);
  paramvalue[4] = &(use_msv = true);
  paramvalue[5] = &(int_method = 1);
  paramvalue[6] = &(permc_spec = 2);
  paramvalue[7] = &(out_steps = 200);
  paramvalue[8] = &(gcomp = 1e-4);
  paramvalue[9] = &(opt = false);
}


SVTran2::~SVTran2()
{
  assert(!superLU);
}

void SVTran2::run(Circuit* cir)
{
  this->cir = cir;
  char msg[80];

  // Build time domain MNAM
  ElemFlag mnam_mask(LINEAR);
  TimeMNAM mnam(cir, mnam_mask);

  // Build T matrix and nonlinear element vector
  n_states = 0;
  int max_n_states = 0;
  ElemFlag mask(NONLINEAR);
  buildTIncidence(mask, cir, T, elem_vec, n_states, max_n_states);

  // Add the compensation resistors to the MNAM
  if (gcomp)
    for (int i = 0; i < n_states; i++)
    {
      mnam.setMAdmittance(T(0,i), T(1,i) , gcomp);
    }
  
  mask = CONVOLUTION;
  // Loop throw all selected elements and build convolution vector if
  // needed.
  ElementVector conv_elem_vec;
  // Reserve a fixed amount since we do not know in advance
  elem_vec.reserve(100);
  cir->setFirstElement(mask);
  Element* elem = cir->nextElement();
  while (elem)
  {
    conv_elem_vec.push_back(elem);
    if (elem->getName() == "yee")
      conv_elem_vec[conv_elem_vec.size() - 1]->getTimeInfo(h, tf);
    // get next element pointer
    elem = cir->nextElement();
  }

  // Go through all the nonlinear elements
  int n_elem = elem_vec.size();
  cns.resize(n_elem);
  mtl::set_value(cns, 0);
  nss.resize(n_elem);
  mtl::set_value(nss, 0);
  n_sec_states = 0;
  for (int k=0; k < n_elem; k++)
  {
    cns[k] = elem_vec[k]->getNumberOfStates();
    nss[k] = elem_vec[k]->getNumberOfSecStates();
    n_sec_states += nss[k];
  }
  
  // this vector contains all the time delays specified in the netlist
  std::vector<double> tempTimeDelays;
  tempTimeDelays.push_back(0.0); // set atleast one element

  // Mask to search through nonlinear elements for delays
  mask = NONLINEAR;
  cir->setFirstElement(mask);
  Element* nl_elem = cir->nextElement();

  // Declare a temporary delay vector storage container, and an iterator
  DenseDoubleVector tempDelayVec;
  DenseDoubleVector::iterator iter;

  // Parse the NL elements and push the delay vector contents into vector<double> tempTimeDelays
  while (nl_elem)
  {
    tempDelayVec = nl_elem->getDelayVec();
    for(iter = tempDelayVec.begin(); iter != tempDelayVec.end(); iter++)
      tempTimeDelays.push_back(*iter);
    nl_elem = cir->nextElement();
  }
  
  // this is the max of all the time delays
  double maxTimeDelay = *max_element(tempTimeDelays.begin(), tempTimeDelays.end());

  // Setup simulation variables (use circular vectors)
  ls_size = mnam.getDim();
  int out_size = int((tf - nst) / h + 1.49999);
  
  // list_size is reduced to the bare minimum number of time steps
  // in order to conserve memory in case of long transient simulations.
  // list_size's value is used in the allocation of all the circular 
  // vectors below.
  list_size = int(2 + maxTimeDelay/h);

  cU = new CircVector(list_size, ls_size);
  cX = new CircVector(list_size, n_states);
  cVnl = new CircVector(list_size, n_states);
  cInl = new CircVector(list_size, n_states);
  cY = new CircVector(list_size, n_sec_states);
  DenseDoubleVector x_v(n_states);
  ssv = DenseDoubleVector(n_states);
  DenseDoubleVector sf(ls_size);
  vl1 = DenseDoubleVector(n_states);
  itmp2 = DenseDoubleVector(n_states);
  tmp_v = DenseDoubleVector(ls_size);
  rhs_v = DenseDoubleVector(ls_size);
  rhs_p = &(rhs_v[0]);
  
  // Choose an integration method and combine M and M'
  LIntegMethod *l_im;
  NLIntegMethod *nl_im;
  if (int_method == 1)
  {
    l_im = new LTrapezoidal(cU, &mnam);
    nl_im = new NLTrapezoidal(cX, cY, h);
  }
  else
  {
    l_im = new LEuler(cU, &mnam);
    nl_im = new NLEuler(cX, cY, h);
  }
  // Create Matrix
  DoubleSparseColMatrix M(ls_size, ls_size, 6 * ls_size);
  l_im->buildMd(M, h);

  // Factor MNAM
  superLUFactor(M);

  // Create Msv if requested
  if (use_msv)
    buildMsv();

  // Create the interface class
  tdsv = new TimeDomainSV(nl_im, max_n_states);

  // --- Optional DC analysis setup
  // Initial condition vectors
  DenseDoubleVector X0(n_states), Vnl0(n_states,zero);
  DenseDoubleVector Inl0(n_states,zero), Vl0(n_states,zero);
  if (opt)
  {
    // run a DC analysis
    DC * dca = new DC();
    double res = dca->getBias(cir, LINEAR, NONLINEAR,
                              X0, Vnl0, Inl0, Vl0);
    sprintf(msg, "DC analysis residual = %g", res);
    report(MESSAGE, msg);

    // Initialize variables related to DC simulation
    assert(int(X0.size()) == n_states);

    // Initialize the circular vectors correctly
    for (int j = 0; j < list_size; j++)
    {
      cX->getCurrent() = X0;
      cX->advance();
      cVnl->getCurrent() = Vnl0;
      cVnl->advance();
      cInl->getCurrent() = Inl0;
      cInl->advance();
      for (int i = 0; i < n_states; i++)
        cU->getCurrent()[i] = Vl0[i];
      cU->advance();
    }
    delete dca;
  }

  // Setup nonlinear solver
  NLSInterface* nlsi = new NLSInterface(this, n_states, deriv);

  // Begin main loop: fixed time step by now
  double residual = zero;
  double residual_max = zero;
  double residual_max_since_last_report = zero;
  int rc_since_last_report = 0;

  report(MESSAGE, "--- Starting transient simulation ...\n");

  sprintf(msg, "Number of nonlinear state variables: %d",n_states);
  report(MESSAGE, msg);

  sepLine();
  report(MESSAGE,
  "|  Step  |    Time (s)  |   Residual   |  Recent Max   |     Max      |");
  sepLine();
  double ctime = zero;
  // Set zero initial conditions
  DenseDoubleVector::iterator ii = x_v.begin();
  DenseDoubleVector::iterator iend = x_v.end();
  for (; ii != iend; ++ii)
    *ii = zero;
    
  // Number of convolution-based elements
  int n_c_elem = conv_elem_vec.size();
  n_tsteps = int(tf / h + 0.499999);

  // ctime is the current time
  for (int nt=0; nt <= n_tsteps; nt++)
  {
    ctime = h * nt;
    // Update time in interface class
    tdsv->setTime(&(cX->getCurrent()[0]),
                  &(cVnl->getCurrent()[0]),
                  &(cInl->getCurrent()[0]),
                  nt, ctime, h);

    // Update compressed source vector
    l_im->buildSf(sf, ctime);
    f1(sf, ssv);

    // Solve the nonlinear equation (call the nonlinear solver).
    double this_res = zero;
    int rc = 0;
  
    // If n_states is zero then there is no need for a nonlinear solution
    if(n_states)
    {
      rc = nlsi->solve(&(x_v[0]), this_res);
      residual = this_res * this_res;
      if(residual > residual_max_since_last_report)
      {
        residual_max_since_last_report = residual;
        if(rc) rc_since_last_report=1;
        if(residual > residual_max) residual_max = residual;
      }
    }

    // update linear network
    updateU(sf);

    // Update convolution-based elements
    // Go through all convolution elements and send result vector
    for (int k = 0; k < n_c_elem; k++)
      conv_elem_vec[k]->setLastResult(cU->getCurrent(), ctime);
    // Update previous derivatives in integration method (if required)
    nl_im->store();
    l_im->store();
    
    // Print table line
    if (!(nt % out_steps))
    {
      if (rc)
          sprintf(msg, "| %6d | %e | %e | %e *| %e |", nt, ctime,
          sqrt(residual) ,sqrt(residual_max_since_last_report),
          sqrt(residual_max));
      else
          sprintf(msg, "| %6d | %e | %e | %e  | %e |", nt, ctime,
          sqrt(residual) ,sqrt(residual_max_since_last_report),
          sqrt(residual_max));
      report(MESSAGE, msg);
      rc = 0;
      rc_since_last_report = 0;
      residual_max_since_last_report = 0;
    }
    
    // PROCESS OUTPUT at each time step instead of doing it at the
    // end of the time loop. This is done to prevent the analysis
    // routine from allocating a large chunk memory of memory before
    // the start of the loop.
    
    // Now fill currents and port voltages of time domain devices.
    int current_ns;
    int n_elem = elem_vec.size();   // Number of nonlinear elements
    int i = 0;
    double tmp_u, tmp_x, tmp_i;

    for (int k = 0; k < n_elem; k++)
    {
      current_ns = elem_vec[k]->getNumberOfStates();
      for (int j = 0; j < current_ns; j++)
      {
        // For the current, decompensate while copying.
        tmp_x = cX->getCurrent()[j+i];
        tmp_i = cInl->getCurrent()[j+i];
        tmp_u = cVnl->getCurrent()[j+i];

        elem_vec[k]->getElemData()->setRealX(j, tmp_i);
        elem_vec[k]->getElemData()->setRealI(j, tmp_i);
        elem_vec[k]->getElemData()->setRealU(j, tmp_u);
      }
      i += current_ns;
    }
    
    Terminal * term = NULL;
    cir->setFirstTerminal();
    while ((term = cir->nextTerminal()) && term->getKeepStatus())
    {
      // Get MNAM index
      if (term->getRC())
      {
        // Remember that MNAM indices begin at 1
        i = term->getRC() - 1;
        tmp_u = cU->getCurrent()[i];
      }
      else // This is a reference terminal
        tmp_u = 0.0;
      
      term->getTermData()->setRealV(tmp_u);
    }
    
    // now for linear element currents
    ElemFlag mask = LINEAR;
    // Loop through all selected elements in circuit and fill current
    // vector if needed.
    cir->setFirstElement(mask);
    Element* elem = cir->nextElement();
    unsigned first_eqn, no_eqn;
    while(0) 
    { // A BUG IT SEEMS (MBS) perhaps to do with keep logic
      while (elem && elem->getKeepStatus())
      {
        elem->getExtraRC(first_eqn, no_eqn);
        cout <<  "first_eqn = " << first_eqn << ", no_eqn = " << no_eqn << endl;
        if (first_eqn)
        {
          // Fill current vector(s) in element
          for (unsigned i = 0; i < no_eqn; i++)
          {
            // Remember that MNAM indices begin at 1
            unsigned row = first_eqn-1 + i;
            tmp_u = cU->getCurrent()[row];
            if(tmp_u) 
              elem->getElemData()->setRealI(i, tmp_u);
          }
        }
        // get next element pointer
        elem = cir->nextElement();
      }
    }
    
    // Advance the circular lists for the next step.
    cU->advance();
    cX->advance();
    cY->advance();
    cVnl->advance();
    cInl->advance();
  } // end time loop
  
  residual = sqrt(residual);

  sprintf(msg, "--- Maximum Residual: %g", sqrt(residual_max));
  report(MESSAGE, msg);
  sepLine();
  newLine();

  // Allocate and copy global time vector
  allocTimeV_P(out_size);
  for (int tindex = 0; tindex < out_size; tindex++)
    TimeV_P[tindex] = nst + h*tindex;

  // Erase allocated space
  delete nlsi;
  delete tdsv;
  delete nl_im;
  delete l_im;
  freeSuperLU();
  delete cY;
  delete cInl;
  delete cVnl;
  delete cX;
  delete cU;
  return;
}

void SVTran2::superLUFactor(DoubleSparseColMatrix& M)
{
  // Call freeSuperLU to release memory
  assert(!superLU);
  double *tmp_p = &(tmp_v[0]);

  // Set source vector to zero 
  // all that is needed now is the L and U factors, not the result
  // of the Solve following the factorization
  DenseDoubleVector::iterator ii = rhs_v.begin();
  DenseDoubleVector::iterator iiend = rhs_v.end();
  for (; ii != iiend; ++ii)
    *ii = zero;

  int nrhs = 1;
  elem_val = new double[M.nnz()];
  row_index = new int[M.nnz()];
  col_pointer = new int[ls_size + 1];
  // copy matrix into a separate structure
  DoubleSparseColMatrix::iterator i, iend;
  DoubleSparseColMatrix::OneD::iterator j, jend;
  int count = 0;
  int col_count = 0;
  i = M.begin();
  iend = M.end();
  for (; i != iend; ++i)
  {
    j = (*i).begin();
    jend = (*i).end();
    col_pointer[col_count++] = count;
    for (; j != jend; ++j)
    {
      elem_val[count] = (*j);
      row_index[count] = j.row();
      count++;
    }
  }
  col_pointer[col_count] = count;
  assert(count == M.nnz());
  char msg[80];
  sprintf(msg, "Matrix size = %d", ls_size);
  report(MESSAGE, msg);
  sprintf(msg, "Matrix nnz = %d", count);
  report(MESSAGE, msg);
  dCreate_CompCol_Matrix(&A, ls_size, ls_size, M.nnz(),
                         elem_val, // elem_val pointer
                         row_index, // row_index pointer
                         col_pointer, // col_pointer pointer
                         NC, _D, GE);
  dCreate_Dense_Matrix(&B, ls_size, nrhs, rhs_p, ls_size, DN, _D, GE);
  perm_r = new int[ls_size];
  perm_c = new int[ls_size];

  // Get column permutation vector perm_c[], according to permc_spec:
  //  permc_spec = 0: use the natural ordering
  //  permc_spec = 1: use minimum degree ordering on structure of A'*A
  //  permc_spec = 2: use minimum degree ordering on structure of A'+A
  get_perm_c(permc_spec, &A, perm_c);

  int info;
  
  // Factorize and Solve
  dgssv(&A, perm_c, perm_r, &L, &U, &B, &info);

  if (info)
    report(FATAL, "There was a problem factoring the MNAM.");

  NCformat *Ustore;
  SCformat *Lstore;
  Lstore = (SCformat *) L.Store;
  Ustore = (NCformat *) U.Store;
  fflush(stdout);
  superLU = true;
}

void SVTran2::freeSuperLU()
{
  assert(superLU);
  // Free vectors
  delete [] perm_c;
  delete [] perm_r;
  // Free supermatrix structures
  Destroy_SuperMatrix_Store(&A);
  Destroy_SuperMatrix_Store(&B);
  // Free L and U matrices
  Destroy_SuperNode_Matrix(&L);
  Destroy_CompCol_Matrix(&U);
  delete [] col_pointer;
  delete [] row_index;
  delete [] elem_val;
  superLU = false;
}

void SVTran2::f1(DenseDoubleVector& s1, DenseDoubleVector& v1l)
{
  assert(superLU);
  // copy source vector to RHS before linear system Solve
  mtl::copy(s1, rhs_v);

  StatInit(sp_ienv(1), sp_ienv(2));
  // Solve the RHS with previously computed L and U
  char trans = 'N';

  int info;

  // Just Solve
  dgstrs(&trans, &L, &U, perm_r, perm_c, &B, &info);
  if (info)
    cerr << "Warning: info = " << info << endl;

  // Premultiply RHS by T' 
  for (int k=0; k < n_states; k++)
  {
    v1l[k] = zero;
    if (T(0,k))
    {
      int tmpidx(T(0,k) - 1);
      v1l[k] += rhs_p[tmpidx];
    }
    if (T(1,k))
    {
      int tmpidx((T(1,k) - 1));
      v1l[k] -= rhs_p[tmpidx];
    }
  }
  StatFree();
  return;
}

void SVTran2::buildMsv()
{
  // This Msv is actually -Msv to avoid having to invert the currents sign
  Msv = DoubleDenseRowMatrix(n_states, n_states);
  // Clear input vector
  DenseDoubleVector::iterator ii = tmp_v.begin();
  DenseDoubleVector::iterator iend = tmp_v.end();
  for (; ii != iend; ++ii)
    *ii = zero;

  for (int i=0; i < n_states; i++)
  {
    // Build by columns
    if (T(0,i))
      tmp_v[T(0,i) - 1] -= one;
    if (T(1,i))
      tmp_v[T(1,i) - 1] += one;
    // Solve linear system and reduce
    f1(tmp_v, vl1);
    // Copy result to Msv column
    for (int j=0; j < n_states; j++)
      Msv(j,i) = vl1[j];
    // Clear vector for next iteration
    if (T(0,i))
      tmp_v[T(0,i) - 1] = zero;
    if (T(1,i))
      tmp_v[T(1,i) - 1] = zero;
  }
  return;
}

void SVTran2::func_ev(double* X, double* F)
{
  // Evaluate first the time-domain elements.
  updateVInl(X);
  // Tell tdsv that the first evaluation is already completed.
  tdsv->clearflag();
  // get nonlinear current and voltages vector
  DenseDoubleVector& vnl = cVnl->getCurrent();
  DenseDoubleVector& inl = cInl->getCurrent();

  if (use_msv)
  {
    ExtDoubleVector itmp1(&(inl[0]), n_states);
    ExtDoubleVector vtmp1(&(vnl[0]), n_states);
    mtl::add(itmp1, mtl::scaled(vtmp1, -gcomp), itmp2);
    mtl::mult(Msv, itmp2, vl1);
  }
  else
  {
    // Calculate error function
    DenseDoubleVector::iterator ii = tmp_v.begin();
    DenseDoubleVector::iterator iend = tmp_v.end();
    for (; ii != iend; ++ii)
      *ii = zero;
    for (int i=0; i < n_states; i++)
    {
      double itmp = inl[i] - gcomp * vnl[i];
      if (T(0,i))
        tmp_v[T(0,i) - 1] -= itmp;
      if (T(1,i))
        tmp_v[T(1,i) - 1] += itmp;
    }
    f1(tmp_v, vl1);
  }
  for (int i=0; i < n_states; i++)
    F[i] = ssv[i] + vl1[i] - vnl[i];
}

void SVTran2::jacobian(double* X, DoubleDenseExtColMatrix& J)
{
  // Set current state variable values.
  for (int j=0; j < n_states; j++)
    cX->getCurrent()[j] = X[j];

  // Number of elements
  int n_elem = elem_vec.size();
  DoubleDenseColMatrix& Ju = tdsv->getJu();
  DoubleDenseColMatrix& Ji = tdsv->getJi();

  if (use_msv)
  {
    // Go through all the nonlinear elements
    for (int i=0, j1=0, k=0; k < n_elem; k++)
    {
      // Set base index in interface object
      tdsv->setBase(i, cns[k], j1, nss[k]);

      // Call element evaluation
      elem_vec[k]->deriv_svTran(tdsv);

      for (int l=0; l < n_states; l++)
        for (int j=0; j < cns[k]; j++)
        {
          J(l, i+j) = zero;
          for (int n=0; n < cns[k]; n++)
            J(l, i+j) += Msv(l, i+n) * (Ji(n, j) - gcomp * Ju(n,j));
        }

      for (int j=0; j < cns[k]; j++)
        for (int n=0; n < cns[k]; n++)
          J(i+j, i+n) -= Ju(j, n);
      i += cns[k];
      j1 += nss[k];
    }
  }
  else
  {
    // Go through all the nonlinear elements
    for (int i=0, j1=0, k=0; k < n_elem; k++)
    {
      // Set base index in interface object
      tdsv->setBase(i, cns[k], j1, nss[k]);
      // Call element evaluation
      elem_vec[k]->deriv_svTran(tdsv);

      for (int j=0; j < cns[k]; j++)
      {
        // Evaluate the linear contribution
        DenseDoubleVector::iterator ii = tmp_v.begin();
        DenseDoubleVector::iterator iend = tmp_v.end();
        for (; ii != iend; ++ii)
          *ii = zero;
        for (int n=0; n < cns[k]; n++)
        {
          int ipn = i + n;
          if (T(0,ipn))
            tmp_v[T(0,ipn) - 1] -= (Ji(n, j) - gcomp * Ju(n,j));
          if (T(1,ipn))
            tmp_v[T(1,ipn) - 1] += (Ji(n, j) - gcomp * Ju(n,j));
        }
        f1(tmp_v, vl1);
        for (int n=0; n < n_states; n++)
          J(n, i+j) = vl1[n];
        // add nonlinear contribution
        for (int n=0; n < cns[k]; n++)
        {
          J(i+n, i+j) -= Ju(n, j);
        }
      }
      i += cns[k];
      j1 += nss[k];
    }
  }
}

void SVTran2::updateVInl(double* x_p)
{
  // Set current state variable values.
  for (int j=0; j < n_states; j++)
    cX->getCurrent()[j] = x_p[j];

  // Number of elements
  int n_elem = elem_vec.size();
  // Go through all the nonlinear elements
  for (int i=0, j=0, k=0; k < n_elem; k++)
  {
    // Set base index in interface object
    tdsv->setBase(i, cns[k], j, nss[k]);
    // Call element evaluation
    elem_vec[k]->svTran(tdsv);
    i += cns[k];
    j += nss[k];
  }
}

void SVTran2::updateU(DenseDoubleVector& s1)
{
  assert(s1.size() == tmp_v.size());
  // copy independent source contribution to tmp_v
  mtl::copy(s1, tmp_v);

  // update tmp_v to include contributions
  // from the nonlinear elements
  DenseDoubleVector& inl = cInl->getCurrent();
  DenseDoubleVector& vnl = cVnl->getCurrent();
  for (int i=0; i < n_states; i++)
  {
    double itmp = inl[i] - gcomp * vnl[i];
    if (T(0,i))
      tmp_v[T(0,i) - 1] -= itmp;
    if (T(1,i))
      tmp_v[T(1,i) - 1] += itmp;
  }

  assert(superLU);
  // now copy the updated tmp_v to RHS vector
  // before calling the linear system Solve
  mtl::copy(tmp_v, rhs_v);
  
  StatInit(sp_ienv(1), sp_ienv(2));
  // Solve the entire system using previously computed
  // L and U factors
  char trans = 'N';
  int info;
  // Solve only
  dgstrs(&trans, &L, &U, perm_r, perm_c, &B, &info);
  if (info)
    cerr << "Warning: info = " << info << endl;

  // copy the RHS solution to update the 
  // terminal voltage vector
  ExtDoubleVector u_v(&(cU->getCurrent()[0]), ls_size);
  mtl::copy(rhs_v, u_v);
  
  StatFree();
}

