#include "SVTran4.h"
#include "TimeMNAM.h"
#include "TimeDomainSV.h"
#include "NLSInterface.h"
#include "Euler.h"
#include "Trapezoidal.h"
#include <iostream>

extern int superLU_print_enable ;

extern "C" 
{
	#include "../inout/ftvec.h"
	#include "../inout/report.h"
}

void buildTIncidence(ElemFlag mask, Circuit*& my_circuit,
		     IntDenseColMatrix& T, ElementVector& elem_vec,
		     int& n_states, int& max_n_states);

// Static members
const unsigned SVTran4::n_par = 15;

// Element information
ItemInfo SVTran4::ainfo = 
{
  "SVTran4",
  "State-Variable-Based Time-Marching Transient Analysis with Newton Iterations and SPICE like time step control",
  "Sonali Luniya",
  DEFAULT_ADDRESS
};

// Parameter information
ParmInfo SVTran4::pinfo[] = {
  {"tstop", "Stop time (s)", TR_DOUBLE, true},
  {"tstep", "Maximum time step (s)", TR_DOUBLE, true},
  {"nst", "No save time (s)", TR_DOUBLE, false},
  {"deriv", "Approximate derivatives or use automatic diff.", TR_INT, false},
  {"msv", "Use Msv flag", TR_BOOLEAN, false},
  {"im", "Integration method", TR_INT, false},
  {"savenode", "Save node voltages", TR_BOOLEAN, false},
  {"permc_spec", "Permutation ordering to factor Msv (0, 1 or 2)",  
   TR_INT, false},
  {"out_steps", "Number of steps skipped for output simulation progress",  
   TR_INT, false},
  {"gcomp", "Compensation network conductance (S)", TR_DOUBLE, false},
  {"cm" , "Compensation method ",TR_DOUBLE, false},
  {"reltol", "Relative Error" , TR_DOUBLE, false},
  {"abstol", "Absolute Error" , TR_DOUBLE, false},
  {"max_iter", "Maximum number of iterations", TR_INT,false},
  {"tmin" , "Minimum time step", TR_DOUBLE , false}
  
};


SVTran4::SVTran4() : Analysis(&ainfo, pinfo, n_par), ls_size(0),  
  superLU(false)
{
  // Parameter stuff
  paramvalue[0] = &(tf);
  paramvalue[1] = &(h);
  paramvalue[2] = &(nst = zero);
  paramvalue[3] = &(deriv = 0);
  paramvalue[4] = &(use_msv = true);
  paramvalue[5] = &(int_method = 0);
  paramvalue[6] = &(savenode = true);
  paramvalue[7] = &(permc_spec = 2);
  paramvalue[8] = &(out_steps = 500);
  paramvalue[9] = &(gcomp = 1e-4);
  paramvalue[10] = &(cm=1.0);
  paramvalue[11] = &(reltol=0.01);
  paramvalue[12] = &(abstol=0.0001);
  paramvalue[13] = &(max_iter=2550);
  paramvalue[14] = &(tmin = 1e-16);
}


SVTran4::~SVTran4()
{
  // assert(!superLU);

}
void SVTran4::run(Circuit* cir)
{
  nt=0;
  total_iteration=0;
  scount=0;
  col_count=0;
  mnam_scount=0;
  cmflag=0;
  sparsity=0;
  cktorder=1;
  prev_timestep=0;
  new_timestep=0;
  ptime=0;
  net_timestep=0;
  net_timestep=h;
  if(cm){
  dummy_run(cir);
  superLU=false;
  dummy_run(cir);
  cmflag=0;
  }
  else{
    cmflag=1;
    dummy_run(cir);
  }
}

void SVTran4::dummy_run(Circuit* cir)
{
  //int mynnz=0;
  
  this->cir = cir;
  char msg[80];
  superLU_print_enable = 0;
  double error = zero, maxerror = zero, maxerror1=zero;
  int iteration =0;
  int flag =0;
  ptime=0;
  n_states = 0;
  int max_n_states = 0;
  h=net_timestep;
  h=net_timestep/4.0;
  int firsttime=2;
  double ctime = zero;

  report(MESSAGE, "*** State Variable-Based Transient Analysis using Sparse matrices and Newton Iterations with SPICE like Time step control***");
  sepLine();
  newLine();
  
  // Build time domain MNAM
  ElemFlag mnam_mask(LINEAR);
  TimeMNAM mnam(cir, mnam_mask);
  // Build T matrix and nonlinear element vector
 
  
  ElemFlag mask(NONLINEAR);
  buildTIncidence(mask, cir, T, elem_vec, n_states, max_n_states);
  if(cm){
    if(!cmflag){
      rcomp = DenseDoubleVector(n_states);
      for(int i=0; i <n_states; i++){
	if(gcomp)
	  rcomp[i]=1.0/gcomp;}
    }
  }
  else
    {
      rcomp = DenseDoubleVector(n_states);
      for(int i=0; i <n_states; i++){
	if(gcomp)
	  rcomp[i]=1.0/gcomp;}
    }
  
 
 
  // Add the compensation resistors to the MNAM
  for (int i = 0; i < n_states; i++) {
    if(rcomp[i])
      mnam.setMAdmittance(T(0,i), T(1,i) , 1.0/rcomp[i]);
  }

  // Go through all the nonlinear elements
  int n_elem = elem_vec.size();
  cns.resize(n_elem);
  nss.resize(n_elem);
  n_sec_states = 0;
  for (int k=0; k < n_elem; k++) {
    cns[k] = elem_vec[k]->getNumberOfStates();
    nss[k] = elem_vec[k]->getNumberOfSecStates();
    n_sec_states += nss[k];
  }


  
  // Setup simulation variables (use circular vectors)
  ls_size = mnam.getDim();
  list_size = int(((tf - nst) / h + 3)*20);
  if (savenode)
    cU = new CircVector(list_size, ls_size);
  else
    cU = new CircVector(51, ls_size);
  if(cm){
    if(!cmflag){ //initialize only once
      cX = new CircVector(list_size, n_states);
      cY = new CircVector(list_size, n_sec_states);
    }
  }
  else{
    cX = new CircVector(list_size, n_states); 
    cY = new CircVector(list_size, n_sec_states);
  }
  cVnl = new CircVector(list_size, n_states);
  cInl = new CircVector(list_size, n_states);
  cTime = new CircVector(list_size, 1);
  cTimestep = new CircVector(list_size,1);
  DenseDoubleVector x_v(n_states);
  ssv = DenseDoubleVector(ls_size+n_states);
  DenseDoubleVector sf(ls_size);
  vl1 = DenseDoubleVector(n_states);
  itmp2 = DenseDoubleVector(n_states);
  tmp_v = DenseDoubleVector(ls_size+n_states);
  rhs_v = DenseDoubleVector(ls_size+n_states);
  c_v = DenseDoubleVector(ls_size+n_states);
  r_v = DenseDoubleVector(ls_size+n_states);
  rhs_p = &(rhs_v[0]);
  r = &(r_v[0]);
  c = &(c_v[0]);
 
  DoubleDenseColMatrix Ju(n_states , n_states);
  DoubleDenseColMatrix Ji(n_states , n_states);
  DoubleDenseColMatrix rhs_temp(ls_size, n_states);
  DoubleSparseColMatrix *M ;
  DenseIntVector myindex(ls_size,0);

  //Set initial conditions
  LIntegMethod *l_im;
  NLIntegMethod *nl_im;
  if (int_method == 1) {
    l_im = new LTrapezoidal(cU, &mnam);
    nl_im = new NLTrapezoidal(cX,cY,cU,cTime, h);
  }
  else {
    l_im = new LEuler(cU, &mnam);
    nl_im = new NLEuler(cX,cY,cU,cTime, h);
  }
   DoubleSparseColMatrix M1(ls_size, ls_size);
   l_im->buildMd(M1, h);
  
   
   // Create the interface class
   // Begin main loop: fixed time step by now
   //double residual = zero;
   report(MESSAGE, "--- Starting transient simulation ...\n");
   sepLine();
   report(MESSAGE, "| Step \t| Time (s) \t|");
   sepLine();
   
  
   M = new DoubleSparseColMatrix(ls_size+n_states,ls_size+n_states);
   //Adding MNAM
   double tx =0;
   rhs_Jv_X = new double[n_states];
   elem_val = new double[(ls_size+n_states)*(ls_size+n_states)];
   row_index = new int[(ls_size+n_states)*(ls_size+n_states)];
   col_pointer = new int[ls_size+n_states+1];
   int myT[n_states][ls_size], tmpidx=0;
   for(int i=0 ; i <n_states;i ++)
     for(int j=0; j<ls_size; j++)
       myT[i][j]=0;
   for(int k =0 ; k < n_states ; k++ )	//Adding T
     {
       if(T(0,k)==T(1,k)){
	 T(0,k) =0;
	 T(1,k)=0;
       }
       if (T(0,k))
	{
	  //mynnz++;
	  tmpidx =T(0,k) -1 ;
	  // (*M)(ls_size+k,tmpidx) = 1;
	  myT[k][tmpidx] = 1;
	  myindex[tmpidx]=1;
	}
       if(T(1,k) )
	{
	  //mynnz++;
	  tmpidx =T(1,k) -1 ;          
	  // (*M)(ls_size+k,tmpidx) = -1;
	  myT[k][tmpidx] =-1;
	    myindex[tmpidx]=1;
        }
     }

   
   for (int i =0 ; i <ls_size ; i++) //col index
      {
	col_pointer[col_count++] = scount;
	for (int j =0 ; j < ls_size  ; j++) //row index
	  {
	    tx = M1(j,i);
	    if(tx!=0){
	      //    (*M)(j,i) = M1(j,i);
	      // mynnz++;
             elem_val[scount] = M1(j,i);
             row_index[scount]= j;
	     scount++;
	     // sprintf(msg,"M[%d][%d] = % f ", j , i , tx);
	     //report(MESSAGE,msg);
	    }
	  }
	for(int j=0 ; j < n_states; j++)
	  {
	    if(myT[j][i]!=0)
	      {
		//(*M)(ls_size+j,i) = myT[j][i];
		elem_val[scount] = myT[j][i];
		row_index[scount]= ls_size+j;
		scount++;
		//double tx= myT[j][i];
		//sprintf(msg,"M[%d][%d] = % f ",j+ls_size,i, tx);
		//report(MESSAGE,msg); 
		rhs_Jv_X[j]=0;
		//	mynnz++;
	      }
	  }
       
      }



 double *temp_mnam;
 temp_mnam = new double[scount];    
 memcpy(temp_mnam, elem_val, sizeof(double)*scount);  
 mnam_scount = scount;
 Integ_predX = new double[n_states];  
 Integ_predU = new double[ls_size];
    //sprintf(msg, "Matrix size = %d", ls_size+n_states);
    //report(MESSAGE, msg);
    //sprintf(msg, "Number of state variables = %d ", n_states);
    //report(MESSAGE,msg);

    n_tsteps = int(tf / h + 1);
   
    //  tdsv = new TimeDomainSV(&(cX->getCurrent()[0]),     //dc constructor  
    //		    &(cVnl->getCurrent()[0]),  
    //		    &(cInl->getCurrent()[0]),max_n_states);
    
    tdsv = new TimeDomainSV(&(cX->getCurrent()[0]),   
			    &(cVnl->getCurrent()[0]),  
			    &(cInl->getCurrent()[0]),max_n_states,nl_im);
    if(!cmflag){
    mtl::set_value(cX ->getCurrent(),zero);
    cX ->advance();
    mtl::set_value(cX ->getCurrent(),zero);
    cX -> advance();
    mtl::set_value(cY ->getCurrent(),zero);
    cY ->advance();
    mtl::set_value(cY ->getCurrent(),zero);
    cY -> advance();

    }
    cTime ->getCurrent()[0]=-h;
    cTime->advance();
    cTime ->getCurrent()[0]=zero;
    cTime ->advance();

    cTimestep->getCurrent()[0]=h;
    cTimestep->advance();
    ctime=h;
    prev_timestep=h;
    while(ptime<tf)
    {
      // nt++;     
      int changestep=0;
     
       //cout<<"Current time: "<<ctime <<" # of Iterations: "<<iteration<<" Time step: "<<h<<endl;
     
      iteration = 0;
      
	if(prev_timestep!=h)
	  {
	    nl_im->changeStep(h);
	    l_im->buildMd(M1,h);
	    
	    for (int i =0 ; i <ls_size ; i++) //col index
	      {
		int tcount =col_pointer[i];
		for (int j =0 ; j < ls_size  ; j++) //row index
		  {
		    tx = M1(j,i);
		    if(tx!=0){
		      elem_val[tcount] = M1(j,i);
		      row_index[tcount]= j;
		      temp_mnam[tcount]=elem_val[tcount];
		      tcount++;
		    }
		  }
	      }
	  }
	cTime->getCurrent()[0] = ctime;
	cTimestep->getCurrent()[0] = h;
	tdsv->setTime(&(cX->getCurrent()[0]),&(cVnl->getCurrent()[0]),  
		      &(cInl->getCurrent()[0]),nt, ctime, h);//first_time=true
          // Update compressed source vector
	l_im->buildSf(sf, ctime);
	nl_im->predictX(n_states,Integ_predX,cTimestep);
	nl_im->predictU(ls_size,Integ_predU,cTimestep);
	do  //Start Newton Iterations
	  {  
	    
	    memset(elem_val,0,sizeof(double)*((ls_size+n_states)*(ls_size+n_states)));
	    memcpy(elem_val, temp_mnam, sizeof(double)*mnam_scount);
	    scount = mnam_scount;
	    col_count = ls_size;
	    mtl::set_value(rhs_temp,0.0);
	    int ibase =0,jbase=0 ;
	    memset(rhs_Jv_X, 0 , n_states*sizeof(double));
       
	    int n_elem = elem_vec.size();
	    int prev_scount=0;
	    for (int k = 0; k < n_elem; k++)  
	      {
		int c_cns = elem_vec[k]->getNumberOfStates();
		// Set base index in interface object
		tdsv->setBase(ibase, cns[k],jbase,nss[k]);
		//elem_vec[k]->deriv_svTran(tdsv);
		elem_vec[k]->svTran(tdsv);
		//tdsv->clearflag();
		elem_vec[k]->deriv_svTran(tdsv);     // Call element evaluation
	   
		DoubleDenseColMatrix& Ju_elem = tdsv->getJu();
		DoubleDenseColMatrix& Ji_elem = tdsv->getJi();
		
		for(int m=0 ; m<c_cns; m++){ //Ji col counter
		  col_pointer[col_count++]= scount;
		  prev_scount = scount;
		  for (int j=0; j < c_cns; j++) { //T' row counter
		    // double temp =0;
		    int tmpidx=0;
		    if(T(0,j+ibase)&&Ji_elem(j,m))         //multiply with 1
		      {
			// double tx = Ji_elem(j,m);
			tmpidx = T(0,j+ibase)-1;
			int i =0;
			for( i = prev_scount; i< scount;i++)
			  {
			    if(row_index[i]==tmpidx){ //element already exists
			      elem_val[i]+=Ji_elem(j,m);
			      break;}
			    if(row_index[i]>tmpidx)
			      {
				for(int ti = scount ; ti > i ; ti--)
				  {
				    elem_val[ti]= elem_val[ti-1];
				    row_index[ti]=row_index[ti-1];
				  }
				elem_val[i]= Ji_elem(j,m);
				row_index[i]=tmpidx;
				scount++;
				break;
			      }
			  }
			if(i==scount) //add new elem at poistion i
			  {
			    elem_val[i]= Ji_elem(j,m);
			    row_index[i]=tmpidx;
			    scount++;
			  }
		      }
		    if(T(1,j+ibase)&&Ji_elem(j,m))         //multiply with -1
		      {
		   //double tx = Ji_elem(j,m);
		   tmpidx = T(1,j+ibase)-1;
		   int i =0;
		   for( i = prev_scount; i< scount;i++)
		     {
		       if(row_index[i]==tmpidx){//element already exists
			 elem_val[i]-=Ji_elem(j,m);
			 break;}
		       if(row_index[i]>tmpidx)
			 {
			   for(int ti = scount ; ti > i ; ti--)
			     {
			       elem_val[ti]= elem_val[ti-1];
			       row_index[ti]=row_index[ti-1];
			     }
			   elem_val[i]= -Ji_elem(j,m);
			   row_index[i]=tmpidx;
			   scount++;
			   break;
			 }
		     }
		   if(i==scount) //add new elem at poistion i
		     {
		       elem_val[i]= -Ji_elem(j,m);
		       row_index[i]=tmpidx;
		       scount++;
		     }
		 }

	      
	     }
	     for(int i=prev_scount; i <scount;i++)
	       rhs_temp(row_index[i],m+ibase)=elem_val[i];
	     for(int i=0; i < c_cns; i++)
	       {
		
		 elem_val[scount]= -Ju_elem(i,m);
		 row_index[scount]=ls_size+i+ibase;
		 scount++;
		 rhs_Jv_X[i+ibase]+=Ju_elem(i,m)*cX->getCurrent()[m+ibase];
		 
	    }
	     
	   }
	     ibase += cns[k];
	     jbase += nss[k];
	 }
     
       tdsv->clearflag();
       DenseDoubleVector& inl = cInl->getCurrent();
       DenseDoubleVector& xnl = cX->getCurrent();
       DenseDoubleVector tmp_vec1(ls_size);

       ExtDoubleVector itmp1(&(inl[0]), n_states);
       ExtDoubleVector xtmp1(&(xnl[0]), n_states);
       DenseDoubleVector& vnl = cVnl->getCurrent();
	      
     //report(MESSAGE, "--Multiplying T'JiXi... \n");
       multiply(rhs_temp , xtmp1, tmp_vec1);
       DenseDoubleVector  tmp_vec2 (ls_size,0);
 
       for(int k=0 ; k <n_states; k++) //Multiplying T'*Inl
	 { 
	   if(rcomp[k])
	     itmp1[k]= itmp1[k]-vnl[k]/rcomp[k];
	   if(T(0,k)){
	     tmpidx=(T(0,k)) -1 ;
	     tmp_vec2[tmpidx] =tmp_vec2[tmpidx]-itmp1[k];
	   }
	   if(T(1,k) ){
	     tmpidx =(T(1,k)) -1;
	     tmp_vec2[tmpidx] =tmp_vec2[tmpidx]+itmp1[k];
	   }
	   
	 }     
       ExtDoubleVector vtmp1(&(vnl[0]), n_states);
       
       DenseDoubleVector rhs_vec(ls_size + n_states);

       for(int i =0 ;i < ls_size; i++)
	 rhs_vec[i] = sf[i] + tmp_vec2[i]+tmp_vec1[i];
       for(int i = 0 ; i < n_states ; i++)
	 rhs_vec[ls_size+i] = vtmp1[i] - rhs_Jv_X[i]; 
       rhs_p = &(rhs_vec[0]);
       superLUFactor(M,ssv);
       //for(int i=0;i<ls_size+n_states;i++)
	//  cout<<"ssv["<<i<<"] = "<<ssv[i]<<endl;
       error = zero;
      
       flag =0 ;
       
       for(int j= 0 ; j < ls_size ; j++){
       	 error = abs(cU->getCurrent()[j] - ssv[j] );  
	 if(error > abstol){
	   flag = 1;
	   break;}
        }
  		
       if(!flag){
       for(int j =0 ; j < n_states ; j++){
       error = abs(cX->getCurrent()[j] - ssv[ls_size+j] );  
       if(error > abstol){
         flag = 1; //continue Newton iteration
	 break;}
       }
       }
       if(!cmflag){
	 iteration = max_iter;
	 ptime=tf;
	 flag=1;
       }
       iteration++;
       updateVInl(&(ssv[0]));
       freeSuperLU();
  
      }while(iteration < max_iter && flag ==1 );

      total_iteration+=iteration;
      
      int jbase =0,ibase=0;
      int n_elem = elem_vec.size();
      for (int k = 0; k < n_elem; k++)  
	{
	  // int c_cns = elem_vec[k]->getNumberOfStates();
	  // Set base index in interface object
	  tdsv->setBase(ibase,cns[k],jbase,nss[k]);       // Call element evaluation
	  elem_vec[k]->svTran(tdsv);
	  if(!cmflag){
	    elem_vec[k]->nx=0;
	    elem_vec[k]->ndx=0;}
	  ibase += cns[k];
	  jbase+= nss[k];
	}
      prev_timestep=h;
      new_timestep= h;
      if(iteration >= max_iter)
	{
	  sprintf(msg , "Newton Raphson  didn't converge |%f %f\t|",maxerror,maxerror1);
	  report(MESSAGE, msg);
	  new_timestep = h = h/2.0;
	  changestep=1;
	}
      else{ 
	if(!firsttime){
	  PredictNewStep(ssv,&new_timestep);
	  //  h=new_timestep;
	  
	  if(new_timestep>0.85 *prev_timestep){
	    new_timestep=h;
	    cktorder=2;
	    PredictNewStep(ssv,&new_timestep);
	    
	    if(new_timestep<=1.05*prev_timestep)
	      cktorder=1;
	    //accept this iteration
	    h= new_timestep;
	  }
	  else
	    {
	      //reject this iteration
	      changestep=1;
	      h = new_timestep;
	    }
	  

	  if(h <= tmin ) {
	    if(prev_timestep > tmin)
	      h = tmin;
	    else{
	      cout<<"Timestep too small. Try increasing tmin and reltol.."<<endl;
	      exit(1);
	    }
	  }
	  if(h>net_timestep)
	    h=net_timestep;
	}
	if(firsttime)
      firsttime--;
      }
      if(cm && (!cmflag)) //Find rcomp
	{
	  
	  report(MESSAGE,"Printing Ii\n");
	  //for(int i = 0; i < n_states; i ++)
	  //cout<<"I["<<i<<"] = "<<cInl->getCurrent()[i]<<"X["<<i<<"] "<<cX->getCurrent()[i]<<endl;
	  DenseDoubleVector tmp_vec4(ls_size,0);
	  DenseDoubleVector linear_i(ls_size,0);
	  DenseDoubleVector delta_i(ls_size,0);
	  //multiply T' and Inl
	  for(int k=0 ; k <n_states; k++)
	    {
	      rcomp[k]=0;
	      if(T(0,k)){
		int tmpidx=(T(0,k)) -1 ;
		tmp_vec4[tmpidx] =tmp_vec4[tmpidx]+cInl->getCurrent()[k];
	      }
	      if(T(1,k) ){
		int tmpidx =(T(1,k)) -1;
		tmp_vec4[tmpidx] =tmp_vec4[tmpidx]-cInl->getCurrent()[k];
	      }
		
	    }

	  DenseDoubleVector icomp(n_states,0);
	  for(int i=0; i <ls_size; i++)
		{
		  
		  //go row by row to compensate delta
		  if(myindex[i])
		    {
		      double denom =0;
		      //multiply M1 and U
		      for(int j =0; j <ls_size;j++)
			{
			  double tx = M1(i,j);
			  linear_i[i] +=tx * cU->getCurrent()[j];
			}
		      delta_i[i]= linear_i[i]+tmp_vec4[i];
		      DenseIntVector sum_index(n_states,0); 
		      for(int k=0; k <n_states ;k++)
			{ //check which currents are involved
			  double tx = myT[k][i];
			  if(tx && !icomp[k]){
			    denom += tx*cInl->getCurrent()[k];
			    sum_index[k]++;
			  }
			}
		      for(int k=0; k<n_states; k++)
			{
			  if(sum_index[k])
			    {
			      if((delta_i[i]==0)||(cInl->getCurrent()[k]==0))
				rcomp[k]=0;
			      else
				rcomp[k]=(cVnl->getCurrent()[k]*denom)/(delta_i[i]*cInl->getCurrent()[k]);
			      rcomp[k]= abs(rcomp[k]);
			      if(rcomp[k] && (rcomp[k]<100))
			      	rcomp[k]=100;
			      //cout<<"rcomp["<<k<<"] "<<rcomp[k]<<endl;
			      icomp[k]=1;
			    }
			}
	    
		    } 
		  
		
		}
	} //end of compensation calculations
      
 
	// Update previous derivatives in integration method (if required)
     if(!changestep && !firsttime){
	nl_im->store();
	l_im->store();
     }
	
     delete tdsv;
	
	// Advance the circular lists for the next step.
     //After dummy run ,run is started with the values of cX from dummy run
     //as Adolc does not erase the tape. ????

     if(cmflag && !changestep){
       nt++;
     cU->advance();
     cX->advance();
     cY->advance();
     cVnl->advance();
     cInl->advance();
     cTime->advance();
     cTimestep->advance();

     }
     // for(int i = 0; i < ls_size; i++)
     //   cU->getCurrent()[i] = ssv[i];
    
     //set initial cX at next time step = calculated cX at this time step
     
       if(changestep)
	 cX->getCurrent()= cX->getPrevious(1);
       
       else
	 for(int i =0 ; i < n_states ; i++){
	   cX->getCurrent()[i] =ssv[ls_size+i];
	 }
     if(nt>=0)    //transient constructor
       {
	 tdsv = new TimeDomainSV(&(cX->getCurrent()[0]),&(cVnl->getCurrent()[0]),  
				 &(cInl->getCurrent()[0]),max_n_states,nl_im);
	 tdsv->clearflag();
       }
     else
       {          //dc constructor
	 tdsv = new TimeDomainSV(&(cX->getCurrent()[0]),   
				 &(cVnl->getCurrent()[0]),  
				 &(cInl->getCurrent()[0]),max_n_states);
	 tdsv->clearflag();
       } 
     
     // Print table line
     if (!(nt % out_steps)) {
       sprintf(msg, "| %d  \t| %g  \t|", nt, ctime);
       report(MESSAGE, msg);
     }
     if(cmflag)
       {
	 ptime=cTime->getPrevious(1)[0];
	 ctime = ptime + h;
       }
     
    
    } // end time loop
    
    superLU = true;
    if(cmflag)
      doOutput();
    cmflag=1;
    delete nl_im;
    delete l_im;
    //freeSuperLU();
    delete cInl;  
    delete cVnl;
    //delete cX;
    delete cU;
    return;
}



void SVTran4::superLUFactor(DoubleSparseColMatrix*& M, DenseDoubleVector& ssv)
{
  // Call freeSuperLU to release memory
  assert(!superLU);
  double *tmp_p = &(ssv[0]);

  int nrhs = 1;
  
  col_pointer[col_count] = scount;
  dCreate_CompCol_Matrix(&A, n_states+ls_size, n_states+ls_size, scount,  
			 elem_val, // elem_val pointer
			 row_index, // row_index pointer
			 col_pointer, // col_pointer pointer
			 NC, _D, GE);
  // dPrint_CompCol_Matrix("A",&A);
 
  dCreate_Dense_Matrix(&B, n_states+ls_size, nrhs, rhs_p, n_states+ls_size, DN, _D, GE);
  dCreate_Dense_Matrix(&X, n_states+ls_size, nrhs, tmp_p, n_states+ls_size, DN, _D, GE);
  perm_r = new int[n_states+ls_size];
  perm_c = new int[n_states+ls_size];
  
  // dPrint_Dense_Matrix("B",&B);
  /*
   * Get column permutation vector perm_c[], according to permc_spec:
   *   permc_spec = 0: use the natural ordering  
   *   permc_spec = 1: use minimum degree ordering on structure of A'*A
   *   permc_spec = 2: use minimum degree ordering on structure of A'+A
   */    	  
  
  get_perm_c(permc_spec, &A, perm_c);
 
  etree = new int[ls_size+n_states];
  factor_param_t *ftp = NULL;
  char fact, equed, trans, refact;
  fact      = 'E';
  equed     = 'B';
  trans     = 'N';
  refact    = 'N';
  double recip_pivot_growth;
  double rcond;
                            // These are vectors of 1 element (nrhs elements)
  double ferr(0), berr(0);
  mem_usage_t mem_usage;
  int info;
  memset(c, 0, (ls_size+n_states)*sizeof(double));
  memset(r, 0, (ls_size+n_states)*sizeof(double));

  dgssvx(&fact, &trans, &refact, &A, ftp, perm_c, perm_r, etree, &equed, r, c,
	 &L, &U, work, lwork, &B, &X, &recip_pivot_growth, &rcond,  
	 &ferr, &berr, &mem_usage, &info);
 
  //cout << "equed: " << equed << endl;
  //cout << "recip_pivot_growth = " << recip_pivot_growth << endl;
  //sprintf(msg, "1 / Condition number = %g", rcond);
  //report(MESSAGE, msg);
  //  sprintf(msg, "info = %d", info);
  // report(MESSAGE, msg);
  //sprintf(msg, "ferr = %g", ferr);
  //report(MESSAGE, msg);
  // sprintf(msg, "berr = %g", berr);
  // report(MESSAGE, msg);

  if (info) 
	{
    report(MESSAGE,"Try increasing tstep..");
    report(FATAL, "There was a problem factoring the MNAM.");
  }

  NCformat *Ustore;
  SCformat *Lstore;
  Lstore = (SCformat *) L.Store;
  Ustore = (NCformat *) U.Store;
  // sprintf(msg, "No of nonzeros in factor L = %d", Lstore->nnz);
  //report(MESSAGE, msg);
  //sprintf(msg, "No of nonzeros in factor U = %d", Ustore->nnz);
  // report(MESSAGE, msg);
  // sprintf(msg, "No of nonzeros in L+U = %d",  
  // Lstore->nnz + Ustore->nnz - ls_size);
  //report(MESSAGE, msg);
  //sprintf(msg, "L\\U MB %.3f\ttotal MB needed %.3f\texpansions %d",
  //	  mem_usage.for_lu/1e6, mem_usage.total_needed/1e6,
  //mem_usage.expansions);

  //report(MESSAGE, msg);
  

  // dPrint_CompCol_Matrix("A",&A);
  //dPrint_SuperNode_Matrix("L",&L);
  //dPrint_CompCol_Matrix("U",&U);
  
  // dPrint_Dense_Matrix("B",&B);
  // dPrint_Dense_Matrix("X",&X);
  fflush(stdout);


  // This is not used anymore
  // Destroy_SuperMatrix_Store(&X);
  superLU = true;
  sparsity =(double) scount/((n_states+ls_size)*(n_states+ls_size));
  sparsity = sparsity*100;
  
  scount=0;
  col_count =0;
}

void SVTran4::freeSuperLU()
{
  assert(superLU);
  // Free vectors
  delete [] etree;
  delete [] perm_c;
  delete [] perm_r;
  // Free supermatrix structures
  Destroy_SuperMatrix_Store(&A);
  Destroy_SuperMatrix_Store(&B);
  // Free L and U matrices
  if ( lwork >= 0 ) {
    Destroy_SuperNode_Matrix(&L);
    Destroy_CompCol_Matrix(&U);
  }
  // delete [] col_pointer;
  //delete [] row_index;
  //delete [] elem_val;
  superLU = false;
}

void SVTran4::f1(DenseDoubleVector& s1, DenseDoubleVector& v1l)
{
  assert(superLU);
  //  do diag(r)*B
  ele_mult(s1, r_v, rhs_v);
  StatInit(sp_ienv(1), sp_ienv(2));
  // Solve using triangular solve for speed
  char trans = 'N';
  int info;
  dgstrs(&trans, &L, &U, perm_r, perm_c, &B, &info);
  if (info)
    cerr << "Warning: info = " << info << endl;
   
  // Premultiply rhs by T' (also premultiply by diag(c))
  for (int k=0; k < n_states; k++) {
    v1l[k] = zero;
    if (T(0,k)) {
      int tmpidx(T(0,k) - 1);
      v1l[k] += rhs_p[tmpidx] * c[tmpidx];
    }
    if (T(1,k)) {
      int tmpidx((T(1,k) - 1));
      v1l[k] -= rhs_p[tmpidx] * c[tmpidx];
    }
  }
  StatFree();
  return;
}

void SVTran4::multiply(DoubleDenseColMatrix& a , ExtDoubleVector& b , DenseDoubleVector& c)
{
  int imax = a.nrows();
  int jmax = a.ncols();

  for(int i =0 ; i < imax ; i++)
    {
      c[i] =0;
    for(int j=0; j < jmax; j++)
      {
	  c[i] += a(i,j) * b[j];
       }
     }
}
  

void SVTran4::buildMsv()
{
  // This Msv is actually -Msv to avoid having to invert the currents sign
  Msv = DoubleDenseRowMatrix(n_states, n_states);
  // Clear input vector
  DenseDoubleVector::iterator ii = tmp_v.begin();
  DenseDoubleVector::iterator iend = tmp_v.end();
  for (; ii != iend; ++ii)  
    *ii = zero;

  for (int i=0; i < n_states; i++) {
    // Build by columns
    if (T(0,i))  
      tmp_v[T(0,i) - 1] -= one;
    if (T(1,i))  
      tmp_v[T(1,i) - 1] += one;
    // Solve linear system and reduce
    f1(tmp_v, vl1);
    // Copy result to Msv column
    for (int j=0; j < n_states; j++)
      Msv(j,i) = vl1[j];
    // Clear vector for next iteration
    if (T(0,i))  
      tmp_v[T(0,i) - 1] = zero;
    if (T(1,i))  
      tmp_v[T(1,i) - 1] = zero;

  }
  // print_all_matrix(Msv);
  return;
}


void SVTran4::func_ev(double* X, double* F)
{
  // Evaluate first the time-domain elements.
  updateVInl(X);
  // Tell tdsv that the first evaluation is already completed.
  tdsv->clearflag();
  // get nonlinear current and voltages vector
  DenseDoubleVector& vnl = cVnl->getCurrent();
  DenseDoubleVector& inl = cInl->getCurrent();

  if (use_msv) 
	{
    ExtDoubleVector itmp1(&(inl[0]), n_states);
    ExtDoubleVector vtmp1(&(vnl[0]), n_states);
    mtl::add(itmp1, scaled(vtmp1, -gcomp), itmp2);
    mult(Msv, itmp2, vl1);
  }
  else {
    // Calculate error function
    DenseDoubleVector::iterator ii = tmp_v.begin();
    DenseDoubleVector::iterator iend = tmp_v.end();
    for (; ii != iend; ++ii)  
      *ii = zero;
    for (int i=0; i < n_states; i++) {
      double itmp = inl[i] - gcomp * vnl[i];
      if (T(0,i))  
	tmp_v[T(0,i) - 1] -= itmp;
      if (T(1,i))  
	tmp_v[T(1,i) - 1] += itmp;
    }
    f1(tmp_v, vl1);
  }
  for (int i=0; i < n_states; i++)  
    F[i] = ssv[i] + vl1[i] - vnl[i];
}

void SVTran4::jacobian(double* X, DoubleDenseColMatrix& J)
{
  // Set current state variable values.
  for (int j=0; j < n_states; j++)
    cX->getCurrent()[j] = X[j];

  // Number of elements
  int n_elem = elem_vec.size();
  int i = 0;
  DoubleDenseColMatrix& Ju = tdsv->getJu();
  DoubleDenseColMatrix& Ji = tdsv->getJi();

  if (use_msv) {
    // Go through all the nonlinear elements
    for (int k = 0; k < n_elem; k++) {
      int cns = elem_vec[k]->getNumberOfStates();
      // Set base index in interface object
      tdsv->setIBase(i, cns);
      // Call element evaluation
      elem_vec[k]->deriv_svTran(tdsv);

      for (int l=0; l < n_states; l++)
	for (int j=0; j < cns; j++) {
	  J(l, i+j) = zero;
	  for (int n=0; n < cns; n++)  
	    J(l, i+j) += Msv(l, i+n) * (Ji(n, j) - gcomp * Ju(n,j));
	}

      for (int j=0; j < cns; j++)  
	for (int n=0; n < cns; n++)
	  J(i+j, i+n) -= Ju(j, n);

      i += cns;
    }
  }
  else {
    // Go through all the nonlinear elements
    for (int k = 0; k < n_elem; k++) {
      int cns = elem_vec[k]->getNumberOfStates();
      // Set base index in interface object
      tdsv->setIBase(i, cns);
      // Call element evaluation
      elem_vec[k]->deriv_svTran(tdsv);
       
      for (int j=0; j < cns; j++) {
	// Evaluate the linear contribution
	DenseDoubleVector::iterator ii = tmp_v.begin();
	DenseDoubleVector::iterator iend = tmp_v.end();
	for (; ii != iend; ++ii)  
	  *ii = zero;
	for (int n=0; n < cns; n++) {
	  int ipn = i + n;
	  if (T(0,ipn))  
	    tmp_v[T(0,ipn) - 1] -= (Ji(n, j) - gcomp * Ju(n,j));
	  if (T(1,ipn))  
	    tmp_v[T(1,ipn) - 1] += (Ji(n, j) - gcomp * Ju(n,j));
	}
	f1(tmp_v, vl1);
	for (int n=0; n < n_states; n++)  
	  J(n, i+j) = vl1[n];
	// add nonlinear contribution
	for (int n=0; n < cns; n++) {
	  J(i+n, i+j) -= Ju(n, j);
	}
      }
      i += cns;
    }
  }
}

void SVTran4::updateVInl(double* x_p)
{
  for(int j=0 ; j< ls_size; j++)  //Copy new value of U into the Circ Vector
       cU->getCurrent()[j] = x_p[j];
   
  for (int j=0; j < n_states; j++) // Copy new value of X into Circ Vector
    cX->getCurrent()[j] = x_p[ls_size+j];
}

int SVTran4::PredictNewStep(DenseDoubleVector& ssv, double *timestep)
{
  double tol=0, diff=0, tmp=0, timetemp=1.7e+18;
  switch(cktorder){
  case 1:
    for(int i=0; i < ls_size; i++){
      tol = MAX(fabs(Integ_predU[i]),fabs(ssv[i]))*reltol + abstol;
      diff = ssv[i] -Integ_predU[i];
      if(fabs(diff)>=reltol){
	tmp= 6.0 * tol * 2.0 /diff;
	tmp = cTimestep->getCurrent()[0] * sqrt(fabs(tmp));
	timetemp = MIN (timetemp,tmp);
      }
    }
    break;
  case 2:
    for(int i=0; i < ls_size; i++){
      tol = MAX(fabs(Integ_predU[i]),fabs(ssv[i]))*reltol + abstol;
      diff = ssv[i] -Integ_predU[i];
      if(fabs(diff)>=reltol){
	tmp = cTimestep->getCurrent()[0]*6.0*tol*2.0 *
	  (cTimestep->getCurrent()[0]+cTimestep->getPrevious(1)[0])/diff;
	tmp = sqrt(fabs(tmp));
	timetemp = MIN (timetemp,tmp);
      }
    }
    break;
  default:
    break;
  }
  *timestep = MIN(2 * *timestep, timetemp);
  return (1);
}

void SVTran4::updateU(DenseDoubleVector& s1)
{
  mtl::copy(s1, tmp_v);
  DenseDoubleVector& inl = cInl->getCurrent();
  DenseDoubleVector& vnl = cVnl->getCurrent();
  for (int i=0; i < n_states; i++) {
    double itmp = inl[i] - gcomp * vnl[i];
    if (T(0,i))  
      tmp_v[T(0,i) - 1] -= itmp;
    if (T(1,i))  
      tmp_v[T(1,i) - 1] += itmp;
  }
  assert(superLU);
  //  do diag(r)*B
  ele_mult(tmp_v, r_v, rhs_v);
  StatInit(sp_ienv(1), sp_ienv(2));
  // Solve using triangular solve for speed
  char trans = 'N';
  int info;
  dgstrs(&trans, &L, &U, perm_r, perm_c, &B, &info);
  if (info)
    cerr << "Warning: info = " << info << endl;

  ExtDoubleVector u_v(&(cU->getCurrent()[0]), ls_size);
  ele_mult(rhs_v, c_v, u_v);
}

void SVTran4::doOutput()  
{
  // First check if the result matrices contain any data
  assert(cX);
  cout<<"Number of state variables :"<<n_states<<endl;
  cout<<"Size of sparse matrix :"<<n_states+ls_size <<"*"<<n_states+ls_size<<endl;
  cout<<"Sparsity of the matrix :"<< sparsity<<"%"<<endl;
  cout<<"Total Number of Newton Iterations :"<<total_iteration<<endl;
  report(MESSAGE, "--- Writing output vectors ...");

  int out_size;
  if(nt>list_size)
    out_size=list_size;
  else
    out_size=nt;

  // Allocate and copy global time vector
  allocTimeV_P(out_size);
  for (int tindex = 0; tindex < out_size; tindex++)
    TimeV_P[tindex] =cTime->getPrevious(out_size-tindex)[0];

  // Temporary vectors  
  DenseDoubleVector tmp_x(out_size);
  DenseDoubleVector tmp_i(out_size);
  DenseDoubleVector tmp_u(out_size);
   
  // Now fill currents and port voltages of time domain devices.
  int current_ns;
  int n_elem = elem_vec.size();   // Number of elements
  int i=0;
  for (int k=0; k < n_elem; k++) {

    current_ns = elem_vec[k]->getNumberOfStates();
    for(int j=0; j< current_ns; j++) {

      // For the current, decompensate while copying.
      for (int tindex=0; tindex < out_size; tindex++) {
	tmp_x[tindex] = cX->getPrevious(out_size - tindex)[j+i];
	tmp_i[tindex] = cInl->getPrevious(out_size - tindex)[j+i];
	tmp_u[tindex] = cVnl->getPrevious(out_size - tindex)[j+i];
      }
      elem_vec[k]->getElemData()->setRealX(j, tmp_i);
      elem_vec[k]->getElemData()->setRealI(j, tmp_i);
      elem_vec[k]->getElemData()->setRealU(j, tmp_u);
    }
    i += current_ns;
  }

  if (savenode) {
    // For each terminal, assign voltage vector
    Terminal* term = NULL;
    cir->setFirstTerminal();
    while((term = cir->nextTerminal())) {
      // Get MNAM index
      if (term->getRC()) {
	// Remember that MNAM indices begin at 1
	i = term->getRC() - 1;  
	for (int tindex=0; tindex < out_size; tindex++)  
	  tmp_u[tindex] = cU->getPrevious(out_size - tindex)[i];
      }
      else
	// This is a reference terminal
	mtl::set_value(tmp_u, zero);
       
      // Set terminal vector
      term->getTermData()->setRealV(tmp_u);
    }
  }
}
