//
// This implementation uses SuperLU
//
#include "WTLinear.h"
#include "MergedMatrix.h"
#include <time.h>
extern "C" {
	#include "iohb.h"
	#include "../inout/report.h"
}

SPStorage::SPStorage()
: col_pointer(NULL), row_index(NULL), elem_val(NULL), rhs(NULL),
ncols(0), nnz(0)
{

}

SPStorage::~SPStorage()
{
  if (nnz) {
    delete [] rhs;
    delete [] x;
    delete [] elem_val;
  }
  if (ncols) {
    delete [] row_index;
    delete [] col_pointer;
  }
}

void SPStorage::updateSize(const int& ncols, const int& nnz)
{
  // Free any allocated memory
  if (this->nnz) {
    delete [] rhs;
    delete [] x;
    delete [] elem_val;
  }
  if (this->ncols) {
    delete [] row_index;
    delete [] col_pointer;
  }
  this->ncols = ncols;
  this->nnz = nnz;
  // Allocate new memory
  if (ncols) {
    col_pointer = new int[ncols + 1];
    rhs = new double[ncols];
    x = new double[ncols];
  }
  if (nnz) {
    row_index = new int[nnz];
    elem_val = new double[nnz];
  }
}


WTLinear::WTLinear() : ready(false), work(NULL), lwork(0)
{

}


WTLinear::~WTLinear()
{
  if (ready) {
    // Free vectors
    delete [] r;
    delete [] c;
    delete [] etree;
    delete [] perm_c;
    delete [] perm_r;
    // Free supermatrix structures
    Destroy_SuperMatrix_Store(&A);
    Destroy_SuperMatrix_Store(&B);
    Destroy_SuperMatrix_Store(&X);
    // Free L and U matrices
    if ( lwork >= 0 ) {
      Destroy_SuperNode_Matrix(&L);
      Destroy_CompCol_Matrix(&U);
    }
  }
}


void WTLinear::setup(WaveletBase* wb, TimeMNAM* mnam,
const IntDenseColMatrix& T, DenseDoubleVector& icv, int& permc_spec)
{
  assert(!ready);
  // Here begins the fun
  // Build the "big matrix"
  this->wb = wb;
  this->mnam = mnam;
  this->T = T;

  // get transformation matrices
  DoubleSparseColMatrix Wj, Wpj;
  wb->getWJ(Wj);
  wb->getWpJ(Wpj);

  // expansion block size
  eblksize = wb->getSize();
  nsamples = eblksize - 1;

  // Build expansion matrices
  eWj = DoubleSparseColMatrix(eblksize-1, eblksize, Wj.nnz());
  DoubleSparseColMatrix eWpj(eblksize-1, eblksize, Wj.nnz());
  CompressedDoubleVector r0_wj(eblksize);

  DoubleSparseColMatrix::iterator i, iend;
  DoubleSparseColMatrix::OneD::iterator j, jend;
  // Fill eWj
  i = Wj.begin();
  iend = Wj.end();
  for (; i != iend; ++i) {
    j = (*i).begin(); jend = (*i).end();
    for (; j != jend; ++j) {
      // Skip elements in row 0
      if (j.row())
				eWj(j.row()-1, j.column()) = (*j);
      else
				// Store elements of Wj(0, 1:n)
			r0_wj[j.column()] = (*j);
    }
  }
  // Fill eWpj
  i = Wpj.begin();
  iend = Wpj.end();
  for (; i != iend; ++i) {
    j = (*i).begin(); jend = (*i).end();
    for (; j != jend; ++j) {
      // Skip elements in row 0
      if (j.row())
				eWpj(j.row()-1, j.column()) = (*j);
    }
  }

  // Combine expansion transformation matrices
  MergedMatrix mW(eWj, eWpj);

  // Get MNAM
  DoubleSparseColMatrix M, Mp;
  mnam->getMatrices(M, Mp);
  // Combine MNAMs
  MergedMatrix mM(M, Mp);

  // Define some shortcut variables
  int w_ncols = eWj.ncols();
  int w_nrows = eWj.nrows();
  int m_ncols = M.ncols();
  int m_nrows = M.nrows();

  //####################################################################
	#if 0
  for (int i=0; i < w_ncols; i++) {
    for (int j=0; j < w_nrows; j++) {
      float tmp = eWj(i,j);
      if (tmp) {
				//printf("x ");
				printf("%+4.2f", tmp);
      }
      else {
				printf("     ");
      }
    }
    printf("\n");
  }
	#endif
  //####################################################################


  // These are the number of elements for each group of matrix expansions.
  int nnzM, nnzMp, nnzMMp;
  mM.getSizes(nnzM, nnzMp);
  nnzMMp = mM.getNNZ() - nnzM - nnzMp;
  // Total number of nonzeros is
  int spnnz = nnzM * eWj.nnz() + nnzMp * eWpj.nnz() +
	nnzMMp * mW.getNNZ() + r0_wj.nnz() * m_nrows;
  // number of columns
  int spncols = m_ncols * w_ncols;
  // Update size in matrix structure
  spstor.updateSize(spncols, spnnz);

  // Fill matrix
  // Set 0 or 1 indexing (1: fortran indexing)
  const int indexing = 0;
  int elem_index = 0;
  for (int k=0; k < m_ncols; k++) {
    int m_nzrows = mM.setColumn(k);
    bool flag;
    // Iterate for each column of Wj / Wjp
    for (int m = 0; m < w_ncols; m++) {
      mM.resetRow();
      flag = true;
      // set beginning of spstor column
      // Adjust indexing
      spstor.col_pointer[k * w_ncols + m] = elem_index + indexing;
      int w_nzrows = mW.setColumn(m);
      for (int l=0; l < m_nzrows; l++) {
				double m_elem, mp_elem;
				int mrow;
				DoubleSparseColMatrix::iterator i;
				DoubleSparseColMatrix::OneD::iterator j, jend;
				int elemtype = mM.getNext(m_elem, mp_elem, mrow);
				if (flag && mrow >= k && r0_wj[m]) {
					// add extra element
					spstor.elem_val[elem_index] = r0_wj[m];
					spstor.row_index[elem_index] = k * eblksize + indexing;
					elem_index++;
					flag = false;
				}
				switch (elemtype) {
					case 0:
					// expand both Wj and Wjp (use mW)
					mW.resetRow();
					for (int n=0; n < w_nzrows; n++) {
						double w, wp;
						int wrow;
						switch (mW.getNext(w, wp, wrow)) {
							case 0:
							spstor.elem_val[elem_index] = m_elem * w + mp_elem * wp;
							break;
							case 1:
							spstor.elem_val[elem_index] = m_elem * w;
							break;
							case 2:
							spstor.elem_val[elem_index] = mp_elem * wp;
							break;
							default:
							// Control should never reach here
							assert(false);
						}
						// Adjust indexing
						spstor.row_index[elem_index] = mrow * eblksize
						+ wrow + 1 + indexing;
						elem_index++;
					}
					break;
					case 1:
					//expand eWj (use eWj)
					i = eWj.begin() + m;
					j = (*i).begin(); jend = (*i).end();
					for (; j != jend; ++j) {
						spstor.elem_val[elem_index] = m_elem * (*j);
						// Adjust indexing
						spstor.row_index[elem_index] = mrow * eblksize
						+ j.row() + 1 + indexing;
						elem_index++;
					}
					break;
					case 2:
					// expand eWpj (use eWpj)
					i = eWpj.begin() + m;
					j = (*i).begin(); jend = (*i).end();
					for (; j != jend; ++j) {
						spstor.elem_val[elem_index] = m_elem * (*j);
						// Adjust indexing
						spstor.row_index[elem_index] = mrow * eblksize
						+ j.row() + 1 + indexing;
						elem_index++;
					}
					break;
					default:
					// Control should never reach here
					assert(false);
				}
				// Check again in case this is the last element in column
				if (flag && l == m_nzrows-1 && r0_wj[m]) {
					// add extra element
					spstor.elem_val[elem_index] = r0_wj[m];
					spstor.row_index[elem_index] = k * eblksize + indexing;
					elem_index++;
					flag = false;
				}
      }
    }
  }
  // Last element in column vector must point to nnz
  spstor.col_pointer[spncols] = elem_index + indexing;
  assert(elem_index == spnnz);

  // Allocate space for time_vec
  time_vec = DenseDoubleVector(w_nrows);
  svw = DenseDoubleVector(spncols);
  update_svw(zero, icv);

  char msg[80];
  sprintf(msg, "Matrix and vector created. Size: %d, nnz: %d\n",
	spncols, spnnz);
  report(MESSAGE, msg);
  sprintf(msg, "Number of time samples: %d\n", w_nrows);

  //####################################################################
	#if 0
	//    // print source vector
	//    for (int i(0); i<m_nrows; i++) {
		//      cout << "----------------------------" << endl;
		//      for (int j(0); j<w_nrows; j++)
		//        cout << time_vec[j] << "\t" << spstor.rhs[i*w_nrows+j] << endl;
	//    }
	//    // Print Wj
	//    cout << "Wj:" << endl;
	//    for (int i=0; i<w_nrows; i++) {
		//      for (int j=0; j<w_ncols; j++)
		//        if (Wj(i,j))
		//  	cout << "x";
		//        else
		//  	cout << " ";
		//      cout << endl;
	//    }

  // print the big matrix
  cout << "Big matrix:" << endl;
  typedef matrix< double,
	rectangle<>,
	compressed<int, external, index_from_zero>,
	column_major>::type ExtSparseMatrix;

  ExtSparseMatrix BM(spncols, spncols, spnnz,
	spstor.elem_val, spstor.col_pointer, spstor.row_index);
  for (int i=0; i<spncols; i++) {
    for (int j=0; j<spncols; j++)
      if (BM(i,j))
				cout << "x";
		else
			cout << " ";
    cout << endl;
  }

	//    // copy matrix
	//    SparseMatrix BM1(spncols, spncols, spnnz);
	//    copy(BM, BM1);
  exit(1);

	#endif
  //####################################################################

  // ******************************************************************
  // Start superLU
  // ******************************************************************
  // Copy source vector to rhs
  ExtDoubleVector spstorrhsv(spstor.rhs, m_nrows * eblksize);
  mtl::copy(svw, spstorrhsv);
  int nrhs = 1;
  dCreate_CompCol_Matrix(&A, spncols, spncols, spnnz,
	spstor.elem_val, spstor.row_index, spstor.col_pointer,
	NC, _D, GE);

  dCreate_Dense_Matrix(&B, spncols, nrhs, spstor.rhs, spncols, DN, _D, GE);
  dCreate_Dense_Matrix(&X, spncols, nrhs, spstor.x, spncols, DN, _D, GE);
  perm_r = new int[spncols];
  perm_c = new int[spncols];

  /*
	* Get column permutation vector perm_c[], according to permc_spec:
	*   permc_spec = 0: use the natural ordering
	*   permc_spec = 1: use minimum degree ordering on structure of A'*A
	*   permc_spec = 2: use minimum degree ordering on structure of A'+A
	*/
  get_perm_c(permc_spec, &A, perm_c);

  etree = new int[spncols];
  r = new double[spncols];
  c = new double[spncols];
  factor_param_t *ftp = NULL;
  char fact, equed, trans, refact;
  fact      = 'E';
  equed     = 'N';
  trans     = 'N';
  refact    = 'N';
  double recip_pivot_growth;
  double rcond;
  // These are vectors of 1 element (nrhs elements)
  double ferr(0), berr(0);
  mem_usage_t mem_usage;
  int info;
  dgssvx(&fact, &trans, &refact, &A, ftp, perm_c, perm_r, etree, &equed, r, c,
	&L, &U, work, lwork, &B, &X, &recip_pivot_growth, &rcond,
	&ferr, &berr, &mem_usage, &info);
  cout << "equed: " << equed << endl;
  cout << "recip_pivot_growth = " << recip_pivot_growth << endl;
  sprintf(msg, "1 / Condition number = %g", rcond);
  report(MESSAGE, msg);
  sprintf(msg, "info = %d", info);
  report(MESSAGE, msg);
  sprintf(msg, "ferr = %g", ferr);
  report(MESSAGE, msg);
  sprintf(msg, "berr = %g", berr);
  report(MESSAGE, msg);

  NCformat *Ustore;
  SCformat *Lstore;
  Lstore = (SCformat *) L.Store;
  Ustore = (NCformat *) U.Store;
  sprintf(msg, "No of nonzeros in factor L = %d", Lstore->nnz);
  report(MESSAGE, msg);
  sprintf(msg, "No of nonzeros in factor U = %d", Ustore->nnz);
  report(MESSAGE, msg);
  sprintf(msg, "No of nonzeros in L+U = %d",
	Lstore->nnz + Ustore->nnz - spncols);
  report(MESSAGE, msg);
  sprintf(msg, "L\\U MB %.3f\ttotal MB needed %.3f\texpansions %d",
	mem_usage.for_lu/1e6, mem_usage.total_needed/1e6,
	mem_usage.expansions);
  report(MESSAGE, msg);
  fflush(stdout);

	#if 0
  //####################################################################
  // Multiply original matrix and result
  ExtVector x1(spstor.x, spncols);
  ExtVector b1(spstor.rhs, spncols);
  DenseVector b2(spncols);
  DenseVector res(spncols);
  mtl::mult(BM1, x1, b2);
  mtl::add(scaled(b1, -one), b2, res);
  double nb1 = two_norm(b1);
  double nres = two_norm(res);
  cout << "\n\n||b-Ay|| = " << nres << endl;
  cout << "||b|| = " << nb1 << endl;
  cout << "||b-Ay|| / ||b|| = " << nres / nb1 << endl;

  exit(1);
  //####################################################################
	#endif

  // Create bMsv and ssv
  n_states = T.ncols();
  int svsize = n_states * w_nrows;
  // Create the 'compressed' source vector
  DenseDoubleVector ssv(svsize);
  for (int i(0); i < svsize; i++)
    ssv[i] = zero;

  // Premultiply spstor.x by T_1,w' and *invert sign* (new definition of
  // error fuction)
  for (int k=0; k < n_states; k++) {
    DenseDoubleVector stmp1(eblksize);
    for (int l=0; l < eblksize; l++) {
      stmp1[l] = zero;
      if (T(0,k))
				stmp1[l] -= spstor.x[(T(0,k) - 1) * eblksize + l];
      if (T(1,k))
				stmp1[l] += spstor.x[(T(1,k) - 1) * eblksize + l];
    }
    // Pre-transform source vector
    ExtDoubleVector stmp2(&(ssv[k*w_nrows]), w_nrows);
    mtl::mult(eWj, stmp1, stmp2);
  }
  this->ssv = ssv;
  backup_ssv = DenseDoubleVector(svsize);
  mtl::copy(ssv, backup_ssv);
	//    cout << "ssv: \n" ;
	//    for (int i=0; i<svsize; i++)
	//      cout << ssv[i] << "\n";
	//    cout << endl;

  // Create the 'compressed' matrix
  DoubleDenseColMatrix bMsv(svsize, svsize);
  sprintf(msg, "Msv size: %d\n", bMsv.nrows());
  report(MESSAGE, msg);
  for (int i(0); i < svsize; i++)
    for (int j(0); j < svsize; j++)
      bMsv(i,j) = zero;

  // Now we do triangular solve on the factored A
  // Required by dgstrs()
  StatInit(sp_ienv(1), sp_ienv(2));
  // Sweep through all the columns of T_1,w
  for (int i=0; i < n_states; i++) {
    DenseDoubleVector stmp1(eblksize);
    for (int j=1; j < w_ncols; j++) {
      // Set rhs to zero
      for (int k=0; k < spncols; k++)
				spstor.rhs[k] = zero;
      // Build the rhs vector (Do diag(r)*B too)
      if (T(0,i)) {
				int tmpidx((T(0,i) - 1) * eblksize + j);
				spstor.rhs[tmpidx] += r[tmpidx];
      }
      if (T(1,i)) {
				int tmpidx((T(1,i) - 1) * eblksize + j);
				spstor.rhs[tmpidx] -= r[tmpidx];
      }
      // Solve using triangular solve for speed
      dgstrs(&trans, &L, &U, perm_r, perm_c, &B, &info);

      if (info)
				cerr << "Warning: info = " << info << endl;

      // Premultiply tempvec2 by T_1,w' (also premultiply by diag(c))
      for (int k=0; k < n_states; k++) {
				for (int l=0; l < eblksize; l++) {
					stmp1[l] = zero;
					if (T(0,k)) {
						int tmpidx((T(0,k) - 1) * eblksize + l);
						stmp1[l] += spstor.rhs[tmpidx] * c[tmpidx];
					}
					if (T(1,k)) {
						int tmpidx((T(1,k) - 1) * eblksize + l);
						stmp1[l] -= spstor.rhs[tmpidx] * c[tmpidx];
					}
				}
				// Pre-transform bMsv column
				ExtDoubleVector stmp2(&(bMsv(k * w_nrows, i * w_nrows + j - 1)),
				w_nrows);
				mtl::mult(eWj, stmp1, stmp2);
				//  	for (int l=0; l < w_ncols; l++)
				//  	  bMsv(k * w_ncols + l, i * w_ncols + j) = stmp2[l];
      }
    }
  }
  StatFree();

  Msv = bMsv;

  //###################################################################
	#if 0

	//    CODenseMatrix DensWj(w_nrows, w_ncols);
	//    CODenseMatrix MBlock(w_nrows, w_ncols);
	//    scale(DensWj, zero);
	//    copy(Wj, DensWj);

	//    // Post-multiply each block of bMsv and see if the result is sparse
	//    for (int i=0; i < n_states; i++)
	//      for (int j=0; j < n_states; j++) {
		//        scale(MBlock, zero);
		//        mult(Msv.sub_matrix(i*w_ncols, (i+1)*w_ncols,
		//  			  j*w_nrows, (j+1)*w_nrows), DensWj, MBlock);
		//        copy(MBlock, Msv.sub_matrix(i*w_ncols, (i+1)*w_ncols,
		//  				  j*w_nrows, (j+1)*w_nrows));
	//      }

  // Print the resulting matrix
  int count(0);
  double thres(.01);
  if (bMsv.nrows() < 100)
    cout << "Msv: " << endl;
  for (int i=0; i<svsize; i++) {
    for (int j=0; j<svsize; j++)
		if (abs(bMsv(i,j)) > thres) {
			count++;
			if (bMsv.nrows() < 100)
				//	  printf(" %+4.2f", bMsv(i,j));
			cout << "x";
		}
		else
			if (bMsv.nrows() < 100)
				// printf("      ");
	  cout << " ";
    if (bMsv.nrows() < 100)
      cout << endl;
  }
  cout << "Density: " << float(count) / bMsv.nrows() / bMsv.nrows() << endl;

  exit(1); // Stop here. We are just testing

	#endif
  //###################################################################

  ready = true;
}

void WTLinear::update_svw(const double& ts, DenseDoubleVector& icv)
{
  int m_nrows = mnam->getDim();
  assert(eblksize);
  assert(int(svw.size()) == m_nrows * eblksize);
  // Get the source vector
  DenseDoubleVector aux_vec(m_nrows);
  DenseDoubleVector time_vec1;
  wb->getTimePointsVector(time_vec1);
  // Get the source vector at each time sample
  for (int i=1; i < eblksize; i++) {
    // Clear vector
    for (int j=0; j<m_nrows; j++)
      aux_vec[j] = zero;
    time_vec[i-1] = time_vec1[i] + ts;
    mnam->getSource(time_vec[i-1], aux_vec);
    for (int j=0; j<m_nrows; j++)
      svw[j*eblksize+i] = aux_vec[j];
    //    print_vector(aux_vec);
  }
  // Set initial conditions
  for (int j=0; j<m_nrows; j++)
    svw[j*eblksize] = icv[j];
}

void WTLinear::update_ssv()
{
  // Now we do triangular solve on the factored A
  // Required by dgstrs()
  StatInit(sp_ienv(1), sp_ienv(2));
  // Solve using triangular solve for speed
  int info;
  char trans = 'N';
  const int& spncols = spstor.getNCols();
  ExtDoubleVector spstorrhsv(spstor.rhs, spncols);
  // do diag(r)*B
  for (int i=0; i < spncols; i++)
    spstorrhsv[i] = svw[i] * r[i];
  //  copy(svw, spstorrhsv);
  dgstrs(&trans, &L, &U, perm_r, perm_c, &B, &info);
  if (info)
    cerr << "Warning: info = " << info << endl;
  StatFree();

  // Premultiply spstor.x by T_1,w' and *invert sign* (new definition of
  // error fuction) (Also premultiply by diag(c))
  for (int k=0; k < n_states; k++) {
    DenseDoubleVector stmp1(eblksize);
    for (int l=0; l < eblksize; l++) {
      stmp1[l] = zero;
      if (T(0,k)) {
				int tmpidx((T(0,k) - 1) * eblksize + l);
				stmp1[l] -= spstor.rhs[tmpidx] * c[tmpidx];
      }
      if (T(1,k)) {
				int tmpidx((T(1,k) - 1) * eblksize + l);
				stmp1[l] += spstor.rhs[tmpidx] * c[tmpidx];
      }
    }
    // Pre-transform source vector
    ExtDoubleVector stmp2(&(ssv[k*nsamples]), nsamples);
    mtl::mult(eWj, stmp1, stmp2);
  }
  mtl::copy(ssv, backup_ssv);
}

void WTLinear::scaleSsv(const double& factor)
{
  mtl::copy(scaled(backup_ssv, factor), ssv);
}

void WTLinear::getVL(const DenseDoubleVector& iNL, DenseDoubleVector& vL)
{
  assert(ready);
  assert(iNL.size() == vL.size() && vL.size() == Msv.nrows());
  mtl::mult(Msv, iNL, ssv, vL);
  return;
}

void WTLinear::getJvL(DoubleDenseColMatrix& JiNL, DoubleDenseExtColMatrix& Jac)
{
  assert(ready);
  mtl::mult(Msv, JiNL, Jac);
}

void WTLinear::addJvLBlock(DoubleDenseColMatrix& JIp,
DoubleDenseExtColMatrix& Jac, const int& blocksize, const int& offset)
{
  assert(ready);
  //
  // Add contribution of JIp block
  DoubleDenseExtColMatrix JacBlock(&(Jac(0,offset)), Jac.nrows(), blocksize);
  DoubleDenseExtColMatrix MsvBlock(&(Msv(0,offset)), Msv.nrows(), blocksize);
  mtl::mult(MsvBlock, JIp.sub_matrix(0, blocksize, 0, blocksize), JacBlock);
}

void WTLinear::getTimeVector(DenseDoubleVector& timev)
{
  assert(time_vec.size() == timev.size());
  mtl::copy(time_vec, timev);
}


void WTLinear::getU(const DenseDoubleVector& iNL, DenseDoubleVector& uw)
{
  // Premultiply iNL by T_1,w' and *invert sign* (new definition of
  // error fuction)
  // Add source vector
  const int& spncols = spstor.getNCols();
  for (int k=0; k < spncols; k++)
    spstor.rhs[k] = svw[k];
  // Sweep through all the columns of T_1,w
  for (int i=0; i < n_states; i++)
	for (int j=1; j < eblksize; j++) {
		// Add the contribution of iNL
		if (T(0,i))
			spstor.rhs[(T(0,i) - 1) * eblksize + j] -=
	  iNL[i * nsamples + j - 1];
		if (T(1,i))
			spstor.rhs[(T(1,i) - 1) * eblksize + j] +=
	  iNL[i * nsamples + j - 1];
	}
  // do diag(r)*B
  for (int k=0; k < spncols; k++)
    spstor.rhs[k] *= r[k];

  // Now we do triangular solve on the factored A
  // Required by dgstrs()
  StatInit(sp_ienv(1), sp_ienv(2));
  // Solve using triangular solve for speed
  int info;
  char trans = 'N';
  dgstrs(&trans, &L, &U, perm_r, perm_c, &B, &info);
  if (info)
    cerr << "Warning: info = " << info << endl;
  StatFree();

  // premultiply by diag(c)
  for (int k=0; k < spncols; k++)
    spstor.rhs[k] *= c[k];

  // transform result
  int nlvars = mnam->getDim();
  for (int i=0; i < nlvars; i++) {
    ExtDoubleVector rhs(spstor.rhs + i*eblksize, eblksize);
    ExtDoubleVector stmp1(&(uw[i*nsamples]), nsamples);
    mtl::mult(eWj, rhs, stmp1);
  }
}

// From ../inout/init_clean.h
extern char spiceTitle[];
extern char inputFilename[];

void WTLinear::writeHB()
{
  assert(ready);
  cout << "\n Saving expanded MNAM ... ";
  // Write matrix and vector in HB format
  int nrhs = 1;
  // Note we do not write guess
  double *guess = NULL;
  // get the first part of the netlist name
  char filename[100];
  strcpy(filename, inputFilename);
  char *dot = strstr(filename, ".net");
  if (dot)
    dot[0] = 0;
  strcat(filename, ".hb");

  char Type[]= "RRA"; // real or complex
  char Ptrfmt[]="(13I6)";
  char Indfmt[]="(16I5)";
  char Valfmt[]="(3E26.18)";
  char Rhsfmt[]="(3E26.18)";

  writeHB_mat_double(filename,
	spstor.getNCols(), spstor.getNCols(), spstor.getNNZ(),
	spstor.col_pointer, spstor.row_index, spstor.elem_val,
	nrhs, spstor.rhs, guess, spstor.x,
	spiceTitle, "EMNAM", Type,
	Ptrfmt, Indfmt, Valfmt, Rhsfmt, "FNX");

  cout << "Matrix saved.\n" << endl;
}

