#include "WTNonLinear.h"
#include <fstream>

WTNonLinear::WTNonLinear(WaveletBase* wb, ElementVector& elem_vec,
int& n_states, int& max_n_states,
WTLinear* wtl, WTDBase* wtr) :
wb(wb), elem_vec(elem_vec),
n_states(n_states), max_n_states(max_n_states), wtl(wtl), wtr(wtr),
npoints(wb->getSize()-1), sysdim(n_states * npoints),
vNL(sysdim), iNL(sysdim), vL(sysdim)
{
  // Create interface object
  xic = DenseDoubleVector(n_states);
  wtr->getXIC(xic);
  wdsv = new WaveletDomainSV(wb, max_n_states, vNL, iNL, xic);

  x = new double[sysdim];
  // Clear vector
  for (int i(0); i<sysdim; i++)
    x[i]=zero;
  nlsi = new NLSInterface(this, sysdim, 0);
}

WTNonLinear::~WTNonLinear()
{
  delete nlsi;
  delete [] x;
  delete wdsv;
}

int WTNonLinear::solve(const int& nsteps)
{
  if (nsteps > 1)
    // Clear vector
	for (int i(0); i<sysdim; i++)
		x[i]=1e-3;

  int retcod;
  if (nsteps == 1)
	{
    wtr->getXIC(xic);
    // Get good guess values
    for (int i = 0; i < n_states; i++)
		{
      ExtDoubleVector xguessi(x + i * npoints, npoints);
			cout << "i = " << i << "\t";
			mtl::print_vector(xguessi);
			cout << endl;
      wb->getGuess(xic[i], xguessi);
			//        print_vector(xguessi);
			//        cout << "###########################" << endl;
    }
    // Call the nonlinear solver
    double residual;
    retcod = nlsi->solve(x, residual);
    cout << "\nRC = " << retcod << endl;
    cout << "Residual: " << residual << endl;
  }
  else
	{
    // Do source stepping
    for (int j=0; j<nsteps; j++)
		{
      double factor = double(j+1) / nsteps;
      cout << "Factor: " << factor << endl;
      wtr->getXIC(xic);
      mtl::scale(xic, factor);
      wtl->scaleSsv(factor);
      // Call the nonlinear solver
      double residual;
      retcod = nlsi->solve(x, residual);
      cout << "\nRC = " << retcod << endl;
      cout << "Residual: " << residual << endl;
      if (residual > 10.)
				// Clear vector and see if next step is better
			for (int i(0); i<sysdim; i++)
				x[i]=zero;
      if (residual < 1e-2)
				// give another chance
			retcod = 0;
    }
  }
  // Original
	//    wtr->getXIC(xic);
	//    // Call the nonlinear solver
	//    double residual;
	//    int retcod = nlsi->solve(x, residual);
	//    cout << "\nRet. cod.: " << retcod << endl;
	//    cout << "Residual: " << residual << endl;

  // Calculate iNL and vNL again
  ExtDoubleVector xv(x, sysdim);
  getVI(xv);

  // Convert x to time doamin before saving
  wtr->getXIC(xic);
  for (int i(0); i < n_states; i++)
	{
    DenseDoubleVector aux(npoints);
    ExtDoubleVector xtemp(x + i * npoints, npoints);
    wb->toTime(xic[i], xtemp, aux);
    mtl::copy(aux, xtemp);
  }

  // Save results
  wtr->saveWindow(wtl, xv, vNL, iNL);

  return retcod;
}

void WTNonLinear::func_ev(double* X, double* F)
{
  // This routine evaluates the following error function
  // F(x) = vNL(x) - vL(x)
  // vL(x) = sSV,w + MSV,w (-iNL(x))

  cout << ".";
  ExtDoubleVector xv(X, sysdim);
  ExtDoubleVector fv(F, sysdim);
  // Calculate iNL and vNL first
  getVI(xv);
  // Get vL
  wtl->getVL(iNL, vL);
  // F = vL - vNL
  mtl::add(vNL, vL, fv);
}

void WTNonLinear::getVI(ExtDoubleVector& xv)
{
  wdsv->setX(xv);
  // Number of elements
  int n_elem = elem_vec.size();
  // Go through all the nonlinear elements
  for (int k(0), i(0); k < n_elem; k++) {
    int cns = elem_vec[k]->getNumberOfStates();
    // Set base index in interface object
    wdsv->setBase(i, cns);
    // Call element evaluation
    elem_vec[k]->svWav(wdsv);
    i += cns;
  }
}


void WTNonLinear::jacobian(double* X, DoubleDenseExtColMatrix& J)
{
  cout << "o";
  cout.flush();

  ExtDoubleVector xv(X, sysdim);
  DoubleDenseExtColMatrix Jac(&(J(0,0)), sysdim, sysdim);
  // Clean Jacobian Matrix
  for (int i(0); i < sysdim; i++)
    for (int j(0); j < sysdim; j++)
      Jac(i,j) = zero;

  // Calculate JacVNL and JacINL
  wdsv->setX(xv);
  // Number of elements
  int n_elem = elem_vec.size();
  // Go through all the nonlinear elements
  for (int k(0), i(0); k < n_elem; k++)
	{
    int cns = elem_vec[k]->getNumberOfStates();
    // Set base index in interface object
    wdsv->setBase(i, cns);
    // Call element evaluation
    elem_vec[k]->deriv_svWav(wdsv);
    int blocksize = cns * npoints;
    int offset = i * npoints;
    // Set contribution of JacVNL
    mtl::copy(wdsv->getJacVp().sub_matrix(0, blocksize, 0, blocksize),
		Jac.sub_matrix(offset, offset + blocksize,
		offset, offset + blocksize));

    // Add contribution of the linear part
    wtl->addJvLBlock(wdsv->getJacIp(), Jac, blocksize, offset);
    i += cns;
  }
}

