#include "WavTran.h"
#include "TimeMNAM.h"
#include "WTLinear.h"
#include "WTNonLinear.h"

extern "C"
{
	#include "../inout/ftvec.h"
	#include "../inout/report.h"
}

void buildTIncidence(ElemFlag mask, Circuit*& my_circuit,
IntDenseColMatrix& T, ElementVector& elem_vec,
int& n_states, int& max_n_states);

// Static members
const unsigned WavTran::n_par = 9;

// Element information
ItemInfo WavTran::ainfo =
{
  "WavTran",
  "Wavelet State Variable Transient Analysis (experimental)",
  "Carlos E. Christoffersen",
  DEFAULT_ADDRESS
};

// Parameter information
ParmInfo WavTran::pinfo[] =
{
  {"l", "L from wavelet method", TR_INT, false},
  {"j", "J from wavelet method", TR_INT, false},
  {"tf", "Final time (s)", TR_DOUBLE, true},
  {"tw", "Initial time window width (s)", TR_DOUBLE, false},
  {"max_nw", "Maximum number of time windows", TR_INT, false},
  {"savemnam", "Flag to save the expanded MNAM", TR_BOOLEAN, false},
  {"nsteps", "Initial number of source steps", TR_INT, false},
  {"stop", "Stop if RC is not zero", TR_BOOLEAN, false},
  {"permc_spec", "Permutation ordering to factor Msv (0, 1 or 2)",
	TR_INT, false}
};


WavTran::WavTran() : Analysis(&ainfo, pinfo, n_par)
{
  // Parameter stuff
  paramvalue[0] = &(L = 6);
  paramvalue[1] = &(J = 0);
  paramvalue[2] = &(tf);
  paramvalue[3] = &(tw);
  paramvalue[4] = &(max_nw = 1);
  paramvalue[5] = &(saveMNAM = false);
  paramvalue[6] = &(nsteps = 1);
  paramvalue[7] = &(stop = true);
  paramvalue[8] = &(permc_spec = 1);
}


WavTran::~WavTran()
{
  // By now it seems that nothing needs to be done here.

}

void WavTran::run(Circuit* cir)
{
  this->cir = cir;

  // Build time domain MNAM
  ElemFlag mnam_mask(LINEAR);
  TimeMNAM mnam(cir, mnam_mask);

  // Build T matrix and nonlinear element vector
  IntDenseColMatrix T;
  ElementVector elem_vec;
  int n_states(0), max_n_states(0);
  ElemFlag mask(NONLINEAR);
  buildTIncidence(mask, cir, T, elem_vec, n_states, max_n_states);

  if (!isSet(&tw)) {
    if (max_nw > 0)
      tw = tf / max_nw;
  }
  if (max_nw < 1)
    report(FATAL, "*** max_nw must be greater than 1.\n");
  if (tw > tf)
    report(FATAL, "*** tw is greater than tf.\n");

  int nw = int(tf / tw + .5);
  if (nw > max_nw)
    max_nw = nw;

  // Build result object
  WTDBase wtr(elem_vec, n_states, max_nw);

  char msg[80];
  sprintf(msg, "\nTime Window %g seconds\n", tw);
  report(MESSAGE, msg);
  double timestart = zero;

  // Build wavelet Base object
  WaveletBase wb(L, J);
  wb.setup(tw);
  int lastsample = wb.getLastIndex() - 1;
  wtr.setWindow(0, wb.getSize() - 1, lastsample);

  int nstart = 0;
  int nstop = nw;

  bool finish = false;
  while (!finish && nstop <= max_nw)
	{
    // Update base matrices
    wb.setTimeWindow(tw);
    // Build linear system
    WTLinear *wtl = new WTLinear();
    DenseDoubleVector u_ic(mnam.getDim());
    wtr.getUIC(u_ic);
    if (permc_spec > 2 || permc_spec < 0)
      permc_spec = 1;
    wtl->setup(&wb, &mnam, T, u_ic, permc_spec);
    if (saveMNAM)
      // Write MNAM in HB format
		wtl->writeHB();

    // Build the nonlinear system
    WTNonLinear *wtnl =
		new WTNonLinear(&wb, elem_vec, n_states, max_n_states, wtl, &wtr);

    // Solve for all time windows
    for (int n = nstart; n < nstop; n++)
		{
      cout << "\nTime window number: "
			<< n << " -- Time: " << timestart << " s" << endl;
      wtr.setWindow(n, wb.getSize() - 1, lastsample);
      wtr.getUIC(u_ic);
      wtl->setTWStart(timestart, u_ic); // Check this !!!
      int rc = wtnl->solve(1);
      if (rc && nsteps > 1)
			{
				// try using source and ic stepping
				rc = wtnl->solve(nsteps);
      }
      if (rc)
			{
				if (stop)
				{
					// Things are not converging. Stop here.
					report(WARNING, "*** No convergence. Giving up.\n");
					finish = true;
					break;
				}
				else
				{
					// Try with a smaller time window
					cout << "Trying smaller window \n";
					sepLine();
					tw = tw / 2.;
					sprintf(msg, "\nTime Window %g seconds\n", tw);
					report(MESSAGE, msg);
					nstart = n;
					nstop = (nstop - n) * 2 + n;
					break;
				}
      }
      else if (n == nstop - 1)
				finish = true;
      // Increment the window start time
      timestart += tw;
    }
    delete wtnl;
    delete wtl;
  }

  // Output results
  wtr.writeOutput();

  return;
}

