#include "WaveletBase.h"
#include <cstdio>

WaveletBase::WaveletBase(const int& L, const int& J)
: L(L), dbL(L), J(J), tw(0), size((1 << (J + 1)) * L + 3), esize(size-1),
ready(false)
{

}

WaveletBase::~WaveletBase()
{

}

void WaveletBase::setup(const double& tw)
{
  assert(!ready);
  this->tw = tw;

  // Use temporary dense matrices to fill the elements
  DoubleDenseColMatrix Wj(size, size);
  DoubleDenseColMatrix Wpj(size, size);
  for (int i(0); i < size; i++)
	for (int j(0); j < size; j++) {
		Wj(i,j) = zero;
		Wpj(i,j) = zero;
	}

  cPoints = DenseDoubleVector(size);
  scPoints = DenseDoubleVector(size);
  unit_xh = DenseDoubleVector(esize);

  // Note: this routine is quite inefficient in generating the
  // transformation matrices. This is because most of the nonzero
  // elements are calculated from the base and wavelet functions even
  // when it is a lot of repetition. Also, the matrices are filled
  // element by element. It would be much faster to build the matrix
  // structure directly.  This approach was chosen to reduce
  // development time. This routine should be replaced to achieve
  // optimum efficiency.

  //#define Wj1(x,y) assert(x < size); assert(y < size); Wj(x,y)

  // Build diagonal blocks and collocation point vector
  // B matrix.
  int L3 = L + 3;
  Wj(0,0) = one;
  Wj(L3 - 1, L3 - 1) = one;
  Wj(1,0) = .125;
  Wj(L3 - 2, L3 - 1) = .125;
  Wj(1,1) = .395833333333333;
  Wj(L3 - 2, L3 - 2) = .395833333333333;
  Wj(1,2) = .260416666666667;
  Wj(L3 - 2, L3 - 3) = .260416666666667;
  Wj(1,3) = .0208333333333333;
  Wj(L3 - 2, L3 - 4) = .0208333333333333;
  Wj(2,1) = .166666666666667;
  Wj(L3 - 3, L3 - 2) = .166666666666667;
  Wj(2,2) = .583333333333333;
  Wj(L3 - 3, L3 - 3) = .583333333333333;
  Wj(2,3) = .166666666666667;
  Wj(L3 - 3, L3 - 4) = .166666666666667;
  for (int i = 3; i < L3-3; i++) {
    Wj(i, i - 1) = .166666666666667;
    Wj(i, i + 1) = .166666666666667;
    Wj(i, i) = .666666666666667;
  }

  //Bp
  Wpj(0,0) = -3.;
  Wpj(L3 - 1, L3 - 1) = 3.;
  Wpj(0,1) = 2.;
  Wpj(L3 - 1, L3 - 2) = -2.;
  Wpj(1,0) = -.75;
  Wpj(L3 - 2, L3 - 1) = .75;
  Wpj(1,1) = -.125;
  Wpj(L3 - 2, L3 - 2) = .125;
  Wpj(1,2) = .8125;
  Wpj(L3 - 2, L3 - 3) = -.8125;
  Wpj(1,3) = .125;
  Wpj(L3 - 2, L3 - 4) = -.125;
  Wpj(2,1) = -.5;
  Wpj(L3 - 3, L3 - 2) = .5;
  Wpj(2,2) = .25;
  Wpj(L3 - 3, L3 - 3) = -.25;
  Wpj(2,3) = .5;
  Wpj(L3 - 3, L3 - 4) = -.5;
  for (int i = 3; i < L3-3; i++) {
    Wpj(i, i - 1) = -.5;
    Wpj(i, i + 1) = .5;
  }

  // X(-1)
  cPoints[0] = zero;
  cPoints[1] = .5;
  for (int i = 2; i < L + 1; i++)
    cPoints[i] = double(i - 1);
  cPoints[L+1] = dbL - .5;
  cPoints[L+2] = dbL;

  // Init Unit vector
  unit_xh[0] = 1.5;
  for (int i(1); i < L; i++)
    unit_xh[i] = one;
  unit_xh[L] = 1.5;
  unit_xh[L+1] = one;
  for (int i(L+2); i < esize; i++)
    unit_xh[i] = zero;

  // Bj and Mj matrices
  int start = L3;
  int nj = 0;
  for (int j = 0; j <= J; j++) {
    start += nj;
    nj = (1 << j) * L;
    int end = start + nj;
    int vsize = (1 << j);

    // X(j)
    cPoints[start] = one / (1 << (j + 2));
    cPoints[end - 1] = L - cPoints[start];
    for (int i = start + 1; i < end - 1; i++)
      cPoints[i] = double(2 * (i - start) + 1) / (1 << (j + 1));

    // Bj
    // Calculate everything, there is not that much repetition if L
    // is small. Besides, the time to generate the whole matrix is
    // quite small.
    for (int i = 0; i < L; i++)
		for (int l = 0; l < vsize; l++) {
			int tmpidx = start + vsize * i + l;
			for (int k = i; k < i + 4; k++) {
				Wj(tmpidx, k) = fik(cPoints[tmpidx], k - 3);
				Wpj(tmpidx, k) = dfik(cPoints[tmpidx], k - 3);
			}
		}

    // M_-1,j
    Wpj(0, start) = dpsijk(cPoints[0], j, -1, nj);
    Wpj(L3-1, start + nj - 1) = - dpsijk(cPoints[0], j, -1, nj);
    int dstart = (j > 0) ? (1<<(j-1)) : 1;
    dstart -= 1;
    Wpj(1, start + dstart) = dpsijk(cPoints[1], j, dstart - 1, nj);
    Wpj(1, start + dstart + 1) =
		dpsijk(cPoints[1], j, dstart, nj);
    Wpj(L3-2, start + nj - dstart - 1) =
		- dpsijk(cPoints[1], j, dstart - 1, nj);
    Wpj(L3-2, start + nj - dstart - 2) =
		- dpsijk(cPoints[1], j, dstart, nj);
    for (int i = 2; i < L3-2; i++) {
      int tmpidx = start + vsize * (i-1) - 1;
      for (int k = 0; k < 2; k++) {
				Wpj(i, tmpidx + k) = dpsijk(cPoints[i], j,
				vsize * (i-1) - 2 + k, nj);
      }
    }

    for (int m = 0; m < j; m++) {
      int vsize1 =  1 << (j-m);
      int njm = L * (1<<m);
      int idxbase = njm + 3;
      // Mjm
      for (int l = 0; l < vsize1; l++) {
				// first and last block (2 elements wide)
				for (int k = 0; k < 2; k++) {
					double tmppsi = psijk(cPoints[start + l], m, k - 1, njm);
					Wj(start + l, idxbase + k) = tmppsi;
					Wj(start + nj - 1 - l, idxbase + njm - 1 - k) = tmppsi;
					tmppsi = dpsijk(cPoints[start + l], m, k - 1, njm);
					Wpj(start + l, idxbase + k) = tmppsi;
					Wpj(start + nj - 1 - l, idxbase + njm - 1 - k) = tmppsi;
				}
				// middle (three elements wide)
				for (int i = 1; i < njm - 1; i++) {
					int tmpidx = start + vsize1 * i + l;
					for (int k = i - 1; k < i + 2; k++) {
						Wj(tmpidx, idxbase + k) =
						psijk(cPoints[tmpidx], m, k - 1, njm);
						Wpj(tmpidx, idxbase + k) =
						dpsijk(cPoints[tmpidx], m, k - 1, njm);
					}
				}
      }

      // Mpmj
      Wpj(idxbase, start) = dpsijk(cPoints[idxbase], j, -1, nj);
      Wpj(idxbase, start + 1) = dpsijk(cPoints[idxbase], j, 0, nj);
      Wpj(idxbase + njm - 1, start + nj - 1) =
			- dpsijk(cPoints[idxbase], j, -1, nj);
      Wpj(idxbase + njm - 1, start + nj - 2) =
			- dpsijk(cPoints[idxbase], j, 0, nj);
      for (int i = 1; i < njm-1; i++) {
				int tmpidx = vsize1 * (i-1) + 3 * (1 << (j-m-1)) -1;
				for (int k = 0; k < 2; k++) {
					Wpj(i + idxbase, start + tmpidx + k) =
					dpsijk(cPoints[i + idxbase], j, tmpidx - 1 + k, nj);
				}
      }


    }

    // Mjj
    Wj(start, start) = psib0(.25); // one;
    Wj(end - 1, end - 1) = psib0(.25); // one;
    Wj(start, start + 1) = .0248618784530387;
    Wj(end - 1, end - 2) = .0248618784530387;
    for (int i = start + 1; i < end - 1; i++) {
      Wj(i, i-1) = .0714285714285714;
      Wj(i, i+1) = .0714285714285714;
      Wj(i, i) = one;
    }
    Wj(start + 1, start) = psib0(1.5); // .0404040404040404;
    Wj(end - 2, end - 1) = psib0(1.5); // .0404040404040404;
    Wj(start + 2, start + 1) = -.0718232044198895;
    Wj(end - 3, end - 2) = -.0718232044198895;

    // Mjjp
    Wpj(start, start) = dpsib0(.25); // -2.30303030303030;
    Wpj(end - 1, end - 1) = -dpsib0(.25); // 2.30303030303030;
    Wpj(start, start + 1) = .0994475138121539;
    Wpj(end - 1, end - 2) = -.0994475138121539;
    for (int i = start + 1; i < end - 1; i++) {
      Wpj(i, i-1) = .428571428571430;
      Wpj(i, i+1) = -.428571428571430;
    }
    Wpj(start + 1, start) = dpsib0(1.5); // -0.242424242424243;
    Wpj(end - 2, end - 1) = -dpsib0(1.5); //0.242424242424243;
    Wpj(start + 1, start + 1) = .033149171270718;
    Wpj(end - 2, end - 2) = -.033149171270718;
    Wpj(start + 2, start + 1) = .430939226519337;
    Wpj(end - 3, end - 2) = -.430939226519337;
  }

  //####################################################################
  // Debug
	#if 0
  for (int i=0; i < size; i++) {
    for (int j=0; j < size; j++) {
      float tmp = Wpj(i,j);
      if (tmp) {
				//printf("x ");
				printf("%+5.3f", tmp);
      }
      else {
				printf("      ");
      }
    }
    printf("\n");
  }
	#endif
  //####################################################################

  // Copy dense matrices into sparse ones
  // Count number of elements
  int countWj(0), countWpj(0);
  for (int i=0; i < size; i++)
	for (int j=0; j < size; j++) {
		if (Wj(j,i))
			countWj++;
		if (Wpj(j,i))
			countWpj++;
	}
  //  cout << "\n" << countWj << "\t" << countWpj << endl;

  this->Wj = DoubleSparseColMatrix(size, size, countWj);
  this->Wpj = DoubleSparseColMatrix(size, size, countWpj);
  Wpsj = DoubleSparseColMatrix(size, size, countWpj);

  // Test if this works
  for (int i=0; i < size; i++)
	for (int j=0; j < size; j++) {
		if (Wj(j,i))
			this->Wj(j,i) = Wj(j,i);
		if (Wpj(j,i))
			this->Wpj(j,i) = Wpj(j,i);
	}

	//    int *colpointer = new int[size+1];
	//    int *rowindex = new int[countWj];
	//    double *elemval = new int[countWj];
	//    int elemindex(0);
	//    for (int i=0; i < size; i++) {
		//      colpointer[i] = elemindex;
		//      for (int j=0; j < size; j++) {
			//        if (Wj(j,i)) {
				//  	rowindex[elemindex] = j;
				//  	elemval[elemindex] = Wj(j,i);
				//  	elemindex++;
			//        }
		//      }
	//    }
	//    colpointer[size] = elemindex;
	//    assert(elemindex == countWj);

	//    typedef matrix< double,
	//      rectangle<>,
	//      compressed<int, external, index_from_zero>,
	//      column_major>::type ExtSparseMatrix;

	//    ExtSparseMatrix WjExt(size, size, countWj, elemval, colpointer, rowindex);

  // Build scaled versions
  double tmp = dbL / tw;
  mtl::copy(scaled(this->Wpj, tmp), Wpsj);
  tmp = tw / dbL;
  mtl::copy(scaled(cPoints, tmp), scPoints);

  // Build reduced matrices
  WJr = DoubleSparseColMatrix(esize, esize, countWj);
  WpsJr = DoubleSparseColMatrix(esize, esize, countWpj);
  c0r_Wj = DenseDoubleVector(esize);
  c0r_Wpj = DenseDoubleVector(esize);

  DoubleSparseColMatrix::iterator i, iend;
  DoubleSparseColMatrix::OneD::iterator j, jend;
  // Fill WJr
  i = this->Wj.begin();
  // Fill c0r_Wj
  for (int k(0); k < esize; k++)
    c0r_Wj[k] = zero;
  j = (*i).begin(); jend = (*i).end();
  for (; j != jend; ++j) {
    // Skip elements in row 0
    if (j.row())
      c0r_Wj[j.row()-1] = (*j);
  }
  // Skip first column for WJr
  i++;
  iend = this->Wj.end();
  for (; i != iend; ++i) {
    j = (*i).begin(); jend = (*i).end();
    for (; j != jend; ++j) {
      // Skip elements in row 0
      if (j.row())
				WJr(j.row()-1, j.column()-1) = (*j);
    }
  }
  // Fill WpsJr
  i = Wpsj.begin();
  // Fill c0r_Wpj
  for (int k(0); k < esize; k++)
    c0r_Wpj[k] = zero;
  j = (*i).begin(); jend = (*i).end();
  for (; j != jend; ++j) {
    // Skip elements in row 0
    if (j.row())
      c0r_Wpj[j.row()-1] = (*j);
  }
  // Skip first column for WpsJr
  i++;
  iend = Wpsj.end();
  for (; i != iend; ++i) {
    j = (*i).begin(); jend = (*i).end();
    for (; j != jend; ++j) {
      // Skip elements in row 0
      if (j.row())
				WpsJr(j.row()-1, j.column()-1) = (*j);
    }
  }

  ready = true;
}


void WaveletBase::setTimeWindow(const double& tw)
{
  double tmp = dbL / tw;
  mtl::copy(scaled(this->Wpj, tmp), Wpsj);
  tmp = tw / dbL;
  mtl::copy(scaled(cPoints, tmp), scPoints);
  // Re-scale reduced matrix
  tmp = this->tw / tw;
  scale(WpsJr, tmp);
  scale(c0r_Wpj, tmp);
  // Update tw value.
  this->tw = tw;
}

void WaveletBase::getTimePointsVector(DenseDoubleVector& result)
{
  assert(ready);
  result = scPoints;
}

void WaveletBase::getWJ(DoubleSparseColMatrix& wj)
{
  assert(ready);
  wj = Wj;
}

void WaveletBase::getWpJ(DoubleSparseColMatrix& wpj)
{
  assert(ready);
  wpj = Wpsj;
}

void WaveletBase::getWJr(DoubleSparseColMatrix& ewj)
{
  assert(ready);
  ewj = WJr;
}


void WaveletBase::getWpJr(DoubleSparseColMatrix& ewpj)
{
  assert(ready);
  ewpj = WpsJr;
}


void WaveletBase::toTime(const double& x0,
const ExtDoubleVector& xhat_r,
DenseDoubleVector& x_r)
{
  assert(ready);
  mtl::mult(WJr, xhat_r, scaled(c0r_Wj, x0), x_r);
  return;
}


void WaveletBase::toTimeDeriv(const double& x0,
const ExtDoubleVector& xhat_r,
DenseDoubleVector& xd_r)
{
  assert(ready);
  mtl::mult(WpsJr, xhat_r, scaled(c0r_Wpj, x0), xd_r);
  return;
}


void WaveletBase::getGuess(const double& x0, ExtDoubleVector& x_r)
{
  mtl::copy(scaled(unit_xh, x0), x_r);
}

//-------------------------------------------------------------------------
//
//               Begin base functions
//
//-------------------------------------------------------------------------
//
// The efficiency in the implementation of the scaling and wavelet
// functions is not very important, at least for the currently planned
// application of this class.

double WaveletBase::fik(const double &x, const int &k)
{
  switch (k) {
		case -3:
    return n1(x);
		case -2:
    return n2(x);
		case -1:
    return phib(x);
		default:
    if (k == L-3)
      return phib(dbL - x);
    else if (k == L-2)
      return n2(dbL - x);
    else if (k == L-1)
      return n1(dbL - x);
  }
  return phi(x - double(k));
}


double WaveletBase::psijk(const double &x, const int& j, const int &k,
double nj)
{
  double j2(1<<j);
  switch (k) {
		case -1:
    return psib0(j2 * x);
		case 0:
    return psib1(j2 * x);
		default:
    if (k == nj-3)
      return psib1(nj - j2 * x);
    else if (k == nj-2)
      return psib0(nj - j2 * x);
  }
  return psi(j2 * x - double(k));
}


double WaveletBase::dfik(const double &x, const int &k)
{
  switch (k) {
		case -3:
    return dn1(x);
		case -2:
    return dn2(x);
		case -1:
    return dphib(x);
		default:
    if (k == L-3)
      return - dphib(dbL - x);
    else if (k == L-2)
      return - dn2(dbL - x);
    else if (k == L-1)
      return - dn1(dbL - x);
  }
  return dphi(x - double(k));
}


double WaveletBase::dpsijk(const double &x, const int& j, const int &k,
double nj)
{
  double j2(1<<j);
  switch (k) {
		case -1:
    return j2 * dpsib0(j2 * x);
		case 0:
    return j2 * dpsib1(j2 * x);
		default:
    if (k == nj-3)
      return - j2 * dpsib1(nj - j2 * x);
    else if (k == nj-2)
      return - j2 * dpsib0(nj - j2 * x);
  }
  return j2 * dpsi(j2 * x - double(k));
}


// Scaling functions
double WaveletBase::n1(const double &x)
{
  // No need to be valid for x<0
  return (x < one) ? pow((one - x), 3) : zero;
}

double WaveletBase::n2(const double &x)
{
  // No need to be valid for x<0
  double res = 2. * x - 3. * x * x + 1.16666666666667 * pow(x, 3.);
  res -= 1.33333333333333 * pow(xp(x - 1), 3.);
  res += 0.166666666666667 * pow(xp(x - 2), 3.);
  return res;
}

double WaveletBase::phib(const double &x)
{
  // No need to be valid for x<0
  double res;
  switch (int(x)+1) {
		case 1:
    res = (1.5-0.91666666666667*x)*x*x;
    break;
		case 2:
    res = (1.5-0.91666666666667*x)*x*x + 1.5 * pow((x-1), 3.);
    break;
		case 3:
    res = (1.5-0.91666666666667*x)*x*x + 1.5 * pow((x-1), 3.)
		- .75 * pow((x-2), 3.);
    break;
		default:
    res = zero;
  }
  return res;
}

double WaveletBase::phi(const double &x)
{
  if (x > zero) {
    double res;
    switch (int(x)+1) {
			case 1:
      res = pow(x, 3.);
      break;
			case 2:
      res = pow(x, 3.) - 4 * pow((x - one), 3.);
      break;
			case 3:
      res = pow(x, 3.) - 4 * pow((x - one), 3.) + 6 * pow((x - 2.), 3.);
      break;
			case 4:
      res = pow(x, 3.) - 4 * pow((x - one), 3.) + 6 * pow((x - 2.), 3.)
			- 4 * pow((x - 3.), 3.);
      break;
			default:
      res = zero;
    }
    return 0.166666666666667 * res;
  }
  else
    return zero;
}

// Wavelet functions
double WaveletBase::psi(const double& x)
{
  double x2 = x * 2.;
  return -0.42857142857143 * phi(x2) + 1.71428571428571 * phi(x2 - one)
	- 0.42857142857143 * phi(x2 - 2.);
}

//  // This is my own function, based on psib()
//  double WaveletBase::psib0(const double& x)
//  {
	//    double x4(x*4.);
	//    return 1.84615384615385 * phib(x4) - 0.46153846153846 * phi(x4);
//  }

double WaveletBase::psib0(const double& x)
{
  return -0.565656565656566 * (psi(x + one) + 14. * psi(x + 2.));
}

double WaveletBase::psib1(const double& x)
{
  return 1.00552486187845 * (psi(x) + 0.0769230769230769 *
	(psi(x + one) + psi(x + 2.)));
}

// Derivatives of scaling and wavelet functions

// Scaling functions (Derivatives)
double WaveletBase::dn1(const double &x)
{
  // No need to be valid for x<0
  return (x < one) ? - 3 * pow((one - x), 2.) : zero;
}

double WaveletBase::dn2(const double &x)
{
  // No need to be valid for x<0
  return 2. - 6. * x + 3.5 * x * x
	- 4. * pow(xp(x-one), 2.) + .5 * pow(xp(x-2), 2.);
}

double WaveletBase::dphib(const double &x)
{
  // No need to be valid for x<0
  double res;
  switch (int(x) + 1) {
		case 1:
    res = (3. - 2.75 * x) * x;
    break;
		case 2:
    res = (3. - 2.75 * x) * x + 4.5 * pow((x - one), 2.);
    break;
		case 3:
    res = (3. - 2.75 * x) * x + 4.5 * pow((x - one), 2.)
		- 2.25 * pow((x - 2), 2.);
    break;
		default:
    res = zero;
  }
  return res;
}

double WaveletBase::dphi(const double &x)
{
  if (x > 0) {
    double res;
    switch (int(x) + 1) {
			case 1:
      res = 3. * x * x;
      break;
			case 2:
      res = 3. * x * x - 12. * pow((x - one), 2.);
      break;
			case 3:
      res = 3. * x * x - 12. * pow((x - one), 2.)
			+ 18. * pow((x - 2.), 2.);
      break;
			case 4:
      res = 3. * x * x - 12. * pow((x - one), 2.)
			+ 18. * pow((x - 2.), 2.) - 12. * pow((x - 3.), 2.);
      break;
			default:
      res = zero;
    }
    return 0.166666666666667 * res;
  }
  else
    return zero;
}

// Wavelet functions (Derivatives)
double WaveletBase::dpsi(const double& x)
{
  double x2 = x * 2.;
  double res = -0.42857142857143 * dphi(x2)
	+ 1.71428571428571 * dphi(x2 - one)
	- 0.42857142857143 * dphi(x2 - 2.);
  return res * 2.;
}

//  // This is my own function, based on psib()
//  double WaveletBase::dpsib0(const double& x)
//  {
	//    double x4(x*4.);
	//    return (1.84615384615385 * dphib(x4) - 0.46153846153846 * dphi(x4)) * 4.;
//  }

double WaveletBase::dpsib0(const double& x)
{
  return -0.565656565656566 * (dpsi(x + one) + 14. * dpsi(x + 2.));
}

double WaveletBase::dpsib1(const double& x)
{
  return 1.00552486187845 * (dpsi(x) + 0.0769230769230769 *
	(dpsi(x + one) + dpsi(x + 2.)));
}

