// Thermal CMOS Inverter
// By Travis Lentz and Tony Mulder

#include <cmath>
#include "../../../../network/ElementManager.h"
#include "../../../../network/AdolcElement.h"
#include "CmosInvT.h"

// Define the number of parameters
const unsigned CmosInvT :: n_par = 23;

// Element Information
ItemInfo CmosInvT::einfo = {
   "cmosinvt",
   "CMOS Inverter, Electro-thermal",
   "Tony Mulder, Travis Lentz",
   DEFAULT_ADDRESS"category:cmos,electrothermal,digital"
   "2005_04_19"
};

// Parameter Information
ParmInfo CmosInvT::pinfo[] = {
   {"vtn","NMOS Threshold Voltage (V)", TR_DOUBLE, false},
   {"vtp","PMOS Threshold Voltage (V)", TR_DOUBLE, false},
   {"un","Effective Mobility of Electrons in NMOS ((cm^2)/(V-sec))", TR_DOUBLE, 
false},
   {"up","Effective Mobility of Holes in PMOS ((cm^2)/(V-sec))", TR_DOUBLE, false},    
   {"en","Permittivity of the Gate Insultor in NMOS (F/cm)", TR_DOUBLE, false},
   {"ep","Permittivity of the Gate Insultor in PMOS (F/cm)", TR_DOUBLE, false},
   {"tox","Thickness of the Gate Insulator (cm)", TR_DOUBLE, false},
   {"wn","Channel Width of NMOS (cm)", TR_DOUBLE, false},
   {"ln","Channel Length of NMOS (cm)", TR_DOUBLE, false}, 
   {"wp","Channel Width of PMOS (cm)", TR_DOUBLE, false},
   {"lp","Channel Length of PMOS (cm)", TR_DOUBLE, false},
   {"td","Response Delay Time (sec)", TR_DOUBLE, false},
   {"thermal", "Thermal element flag", TR_BOOLEAN, false},
   {"ts", "Kirchhoff transformation temperature (K)", TR_DOUBLE, false}, 
   {"tnom", "Ambient temperature (K)", TR_DOUBLE, false},
   {"zt", "Thermal impedence summation (K/W)", TR_DOUBLE, false},
   {"c1", "P channel GS capacitance (F)", TR_DOUBLE, false},
   {"c2", "N channel GS capacitance (F)", TR_DOUBLE, false},
   {"c3", "output GS (Miller) capacitance (F)", TR_DOUBLE, false},
   {"c4", "parasitic diode capacitance from out to Vdd (F)", TR_DOUBLE, false},
   {"c5", "parasitic diode capacitance from out to Gnd(F)", TR_DOUBLE, false},
   {"freq", "operating frequency (Hz)", TR_DOUBLE, false},
   {"lk", "leakage current (A)", TR_DOUBLE, false}
};

// Constructor
CmosInvT::CmosInvT(const string& iname)
   :AdolcElement(&einfo, pinfo, n_par, iname)
{
   // Set default parameter values
   paramvalue[0] = &(vtn = 1);
   paramvalue[1] = &(vtp = -1);
   paramvalue[2] = &(un = 500);
   paramvalue[3] = &(up = 200);
   paramvalue[4] = &(en = 34.515e-14);
   paramvalue[5] = &(ep = 34.515e-14);   
   paramvalue[6] = &(tox = 2e-6);
   paramvalue[7] = &(ln = 2.0e-6);
   paramvalue[8] = &(wn = 50e-6);
   paramvalue[9] = &(lp = 2.0e-6);
   paramvalue[10] = &(wp = 100e-6);
   paramvalue[11] = &(td = 0);
   paramvalue[12] = &(thermal = false);
   paramvalue[13] = &(ts = 300.);
   paramvalue[14] = &(tnom = 300);
   paramvalue[15] = &(zt = 310);
   paramvalue[16] = &(c1 = 1.0e-12);
   paramvalue[17] = &(c2 = 1.0e-12);
   paramvalue[18] = &(c3 = 1.0e-12);
   paramvalue[19] = &(c4 = 1.0e-12);
   paramvalue[20] = &(c5 = 1.0e-12);
   paramvalue[21] = &(freq = 1.0e6);
   paramvalue[22] = &(lk = 8e-6);
 
   // Set Number of Terminals
   setNumTerms(6);

   // Set Flags
   setFlags(NONLINEAR | MULTI_REF | TR_TIME_DOMAIN);

   // Set Number of States
   setNumberOfStates(4);
}

// Initialization Function
void CmosInvT::init() throw(string&)
{
   // Find Beta for the NMOS and PMOS
   betan = (un * en * wn) / (tox * ln);
   betap = (up * ep * wp) / (tox * lp);
   power = ( freq * (c1+c2+c3+c4+c5));

   //create tape
   DenseIntVector var(4);
   var[0] = 0;
   var[1] = 1;
   var[2] = 2;
   var[3] = 3;
   DenseIntVector dvar(4);
   dvar[0] = 0;
   dvar[1] = 1;
   dvar[2] = 2;
   dvar[3] = 3;
   DenseIntVector none(4,0);
   DenseIntVector tvar(4);
   tvar[0] = 0;
   tvar[1] = 1;
   tvar[2] = 2;
   tvar[3] = 3;
   DenseDoubleVector delay(4);
   delay[0] = 0;
   delay[1] = td;
   delay[2] = 0;
   delay[3] = 0;
   createTape(var, dvar, none, tvar, delay);		
}


void CmosInvT::getLocalRefIdx(UnsignedVector& local_ref_vec,
			      TerminalVector& term_list)
{
  // Make sure the vectors are empty
  term_list.erase(term_list.begin(), term_list.end());
  local_ref_vec.erase(local_ref_vec.begin(), local_ref_vec.end());

  // Insert vector elements
  term_list.push_back(getTerminal(0));
  term_list.push_back(getTerminal(1));
  term_list.push_back(getTerminal(2));
  term_list.push_back(getTerminal(3)); // Local reference terminal
  term_list.push_back(getTerminal(4));
  term_list.push_back(getTerminal(5)); // Local reference terminal
  
  local_ref_vec.push_back(3); // Local reference index
  local_ref_vec.push_back(5); // Local reference index
//}
}

// Evaluate Function
void CmosInvT::eval(adoublev& x, adoublev& vp, adoublev& ip)
{
    //x[0]  : Vdd    x[4] : d(Vdd)/dt
    //x[1]  : Vin    x[5] : d(Vin)/dt
    //x[2]  : Vout   x[6] : d(Vout)/dt
    //x[3]  : temp 
    //x[10] : Vout(t-td)
    //vp[0] : Vdd   ip[0] : Isdp
    //vp[1] : Vin   ip[1] : Ig
    //vp[2] : Vout  ip[2] : Iout  ip[3] : pout
    //vp[3] : temp (if thermal is true)
    
    adouble Idsn, Idsp, Vout, VoutTime;
    adouble IdsnB, IdspB, VoutB, tempB;
    adouble IdsnC, IdspC, VoutC, tempC;
    adouble IdsnD, IdspD, VoutD, tempD;
    
    
    //******************* Region A: 0 <= Vin <= Vtn*****************
    Vout = x[0];
    Idsn = lk;
    Idsp = -lk;
          
    // *******************Region B: Vtn <= Vin <= Vdd/2**************
    IdsnB = ((betan/2) * (x[13] - vtn) * (x[13] - vtn))+(power*x[0])+lk;
    IdspB = (-betap * (((x[13] - x[0] - vtp) * (x[2] - x[0])) 
       - ((x[2] - x[0]) * (x[2] - x[0]) / 2)))-(power*x[0])-lk;
    tempB = (x[13] - vtp)*(x[13] - vtp);
    tempB = tempB - (2 * x[0] * (x[13] - (x[0]/2) - vtp));
    tempB = tempB - ((betan/betap) * (x[13] - vtn) * (x[13] - vtn));
    VoutB = (x[13] - vtp) + sqrt(tempB);
    
    condassign(Vout,(x[13]-vtn),VoutB,Vout);
    condassign(Idsn,(x[13]-vtn),IdsnB,Idsn);
    condassign(Idsp,(x[13]-vtn),IdspB,Idsp);
     
    // ****************** Region C: Vin = Vdd/2***********************
    IdsnC = (betan * (x[13] - vtn) * (x[13] - vtn) / 2)+(power*x[0])+lk;
    IdspC = -IdsnC;
    VoutC = x[2];
    
    //condassign(Vout,(x[1]-(x[0]/2)),VoutC,Vout);
    //condassign(Idsn,(x[1]-(x[0]/2)),IdsnC,Idsn);
    //condassign(Idsp,(x[1]-(x[0]/2)),IdspC,Idsp);
    
    // ****************** Region D: Vdd/2 <= Vin <= Vdd + Vtp **********
    IdsnD = (betan * (((x[13] - vtn) * x[2]) - (x[2] * x[2] / 2))) +
    (power*x[0])+lk;
    IdspD = (- (betap/2) * (x[13] - x[0] - vtp) * (x[13] - x[0] - vtp))
      - (power*x[0])-lk;
    tempD = (x[13] - vtn) * (x[13] - vtn);
    tempD = tempD - ((betap/betan) * (x[13] - x[0] - vtp) * (x[13] - x[0] - vtp));
    VoutD = (x[13] - vtn) - sqrt(tempD);
    
    condassign(Vout,(x[13]-(x[0]/2)),VoutD,Vout);
    condassign(Idsn,(x[13]-(x[0]/2)),IdsnD,Idsn);
    condassign(Idsp,(x[13]-(x[0]/2)),IdspD,Idsp);
    
    // ****************** Region E: Vin >= Vdd + Vtp*********************
    condassign(Vout,(x[13]-x[0]-vtp),0,Vout);
    condassign(Idsn,(x[13]-x[0]-vtp),lk,Idsn);
    condassign(Idsp,(x[13]-x[0]-vtp),-lk,Idsp);
    
    
    vp[0] = x[0];
    vp[1] = x[1];
    vp[2] = Vout;
   
    vp[3] =  tnom + x[3]*zt;

    ip[0] = -Idsp;
    ip[1] = 0;
    ip[2] = Idsn + Idsp;

    ip[3] = ((vp[0]-vp[2])*Idsp)-(vp[2]*Idsn)-(power*x[0])-lk;
}
