#include "../../../../network/CircuitManager.h"
#include "../../../../network/AdolcElement.h"
#include "DiodeSP.h"

// Static members
const unsigned DiodeSP::n_par = 12;

// Element information
ItemInfo DiodeSP::einfo =
{
  "diodesp",
  "Spice diode model (conserves charge)",
  "Carlos E. Christoffersen, Nikhil M. Kriplani",
  DEFAULT_ADDRESS"category:diode",
  "2000_07_20"
};

// Parameter information
ParmInfo DiodeSP::pinfo[] =
{
  {"is", "Saturation current (A)", TR_DOUBLE, false},
  {"n", "Emission coefficient", TR_DOUBLE, false},
  {"ibv", "Current magnitude at the reverse breakdown voltage (A)",
	TR_DOUBLE, false},
  {"bv", "Breakdown voltage (V)", TR_DOUBLE, false},
  {"fc", "Coefficient for forward-bias depletion capacitance",
	TR_DOUBLE, false},
  {"cj0", "Zero-bias depletion capacitance (F)", TR_DOUBLE, false},
  {"vj", "Built-in junction potential (V)", TR_DOUBLE, false},
  {"m", "PN junction grading coefficient", TR_DOUBLE, false},
  {"tt", "Transit time (s)", TR_DOUBLE, false},
  {"area", "Area multiplier", TR_DOUBLE, false},
  {"charge", "Use charge-conserving model", TR_BOOLEAN, false},
  {"rs", "Series resistance (ohms)", TR_DOUBLE, false}
};


DiodeSP::DiodeSP(const string& iname) : AdolcElement(&einfo, pinfo, n_par, iname)
{
  // Set default parameter values
  paramvalue[0] = &(is = 1e-14);
  paramvalue[1] = &(n = one);
  paramvalue[2] = &(ibv = 1e-10);
  paramvalue[3] = &(bv = 40.0);
  paramvalue[4] = &(fc = .5);
  paramvalue[5] = &(cj0 = zero);
  paramvalue[6] = &(vj = one);
  paramvalue[7] = &(m = .5);
  paramvalue[8] = &(tt = zero);
  paramvalue[9] = &(area = one);
  paramvalue[10] = &(charge = true);
  paramvalue[11] = &(rs = zero);

  // Set flags
  setFlags(NONLINEAR | ONE_REF | TR_TIME_DOMAIN);
}

void DiodeSP::init() throw(string&)
{
  if (charge)
	{
    // Add one terminal
    Circuit* cir = getCircuit();
    unsigned tref_id = getTerminal(1)->getId();
    unsigned term_id1 = cir->addTerminal(getInstanceName()
		+ ":extra");
    cir->connect(getId(), term_id1);
    // Connect an external 1K resistor (not really needed if using the
    // augmentation network)
    unsigned newelem_id =
		cir->addElement("resistor", getInstanceName() + ":res");
    cir->connect(newelem_id, tref_id);
    cir->connect(newelem_id, term_id1);
    Element* elem = cir->getElement(newelem_id);
    double res = 1e3;
    elem->setParam("r", &res, TR_DOUBLE);
    elem->init();

    // Set the number of terminals
    setNumTerms(3);
    // Set number of states
    setNumberOfStates(2);

    // create tape
    DenseIntVector var(2);
    var[0] = 0;
    var[1] = 1;
    DenseIntVector dvar(1,1);
    createTape(var, dvar);
  }
  else
	{
    // Set the number of terminals
    setNumTerms(2);
    // Set number of states
    setNumberOfStates(1);

    // create tape
    DenseIntVector var(1,0);
    createTape(var, var);
  }
}

void DiodeSP::eval(adoublev& x,
adoublev& vp, adoublev& ip)
{
  double alfa = eCharge / n / kBoltzman / 300.; // tnom = 300K
  double v1 = log(5e8 / alfa) / alfa; // normal is .5e9
  double k3 = exp(alfa * v1);

  // x[0]: x as in Rizolli's equations
  adouble vd, id, ibreakdown;
  // Diode voltage
  condassign(vd, v1 - x[0],
	x[0] + zero,
	v1 + log(one + alfa*(x[0]-v1))/alfa);

  // Static current
  condassign(id, v1 - x[0],
	is * (exp(alfa * x[0]) - one),
	is * k3 * (one + alfa * (x[0] - v1)) - is);

  // breakdown current (taken from Antognetti's SPICE book)
  if (vd == -bv)
    ibreakdown = -ibv;
  else if (vd < -bv)
    ibreakdown = -is * exp(-alfa*(vd + bv) - 1.0 + bv/alfa);
  else
    ibreakdown = 0.0;

  // include the breakdown current contribution
  ip[0] = id + ibreakdown;

  if (charge)
	{
    // x[1]: q
    // x[2]: dq/dt
    // Form the additional error function
    adouble qvj;
    double km;
    if (isSet(&cj0))
		{
      condassign(qvj, fc * vj - vd,
			vj * cj0 / (one - m) * (one - pow(one - vd / vj, one - m)),
			cj0 * pow(one - fc, - m - one) *
			(((one - fc * (one + m)) * vd + .5 * m * vd * vd / vj) -
		  ((one - fc * (one + m)) * vj * fc + .5 * m * vj * fc * fc)) +
			vj * cj0 / (one - m) * (one - pow(one - fc, one - m)));
      km = vj * cj0 / (one - m) * 1e1;
    }
    else
		{
      qvj = zero;
      km = 1e-12;
    }
    if (isSet(&tt))
		{
      qvj += tt * ip[0];
      km += tt * 1e-2;
    }
    // Add capacitor current
    ip[0] += x[2] * km;
    ip[1] = - ip[0];
    vp[1] = qvj / km - x[1];
    vp[0] = vd + ip[0] * rs + vp[1];

    // scale the current according to area.
    ip[0] *= area;
    ip[1] *= area;
  }
  else
	{
    adouble dvd_dx;
    condassign(dvd_dx, v1 - x[0],
		one,
		one / (one + alfa*(x[0]-v1)));
    adouble cd;
    if (isSet(&cj0))
      condassign(cd, fc * vj - vd,
		cj0 * pow(one - vd / vj, -m),
		cj0 * pow(one - fc, - m - one) *
		((one - fc * (one + m)) + m * vd / vj));
    else
      cd = zero;

    if (isSet(&tt))
      cd += alfa * tt * ip[0];
    // Add capacitor current
    ip[0] += cd * dvd_dx * x[1];
    vp[0] = vd + ip[0] * rs;
    // scale the current according to area.
    ip[0] *= area;
  }
}

