#include "../../../../network/ElementManager.h"
#include "../../../../network/AdolcElement.h"
#include "ThermalBlockBC1.h"

// Static members
const unsigned ThermalBlockBC1::n_par = 4;
const double ThermalBlockBC1::sigma = 5.67051e-8;

// Element information
ItemInfo ThermalBlockBC1::einfo =
{
  "thermalblockbc1",
  "Thermal nonlinear boundary condition element",
  "Carlos E. Christoffersen, Bill Batty",
  DEFAULT_ADDRESS"category:thermal",
  "2001_10_10"
};

// Parameter information
ParmInfo ThermalBlockBC1::pinfo[] =
{
  {"epsilon", "Surface emissivity", TR_DOUBLE, false},
  {"hn", "Natural convection heat transfer coefficient", TR_DOUBLE, false},
  {"hf", "Forced convection heat transfer coefficient", TR_DOUBLE, false},
  {"tambient", "Ambient temperature (K)", TR_DOUBLE, false}
};

ThermalBlockBC1::ThermalBlockBC1(const string& iname)
: AdolcElement(&einfo, pinfo, n_par, iname)
{
  // Set default parameter values
  paramvalue[0] = &(epsilon = 0.7);
  paramvalue[1] = &(hn = zero);
  paramvalue[2] = &(hf = zero);
  paramvalue[3] = &(T0 = 300.);

  // Set the number of terminals
  setNumTerms(2);

  // Set flags
  setFlags(NONLINEAR | ONE_REF | TR_TIME_DOMAIN);

  // Set number of states
  setNumberOfStates(1);
}

void ThermalBlockBC1::init() throw(string&)
{
  if (!isSet(&hf))
    hf = 1.3*(1.0+3.)*4.0*T0*T0*T0*sigma*epsilon;
  // create tape
  DenseIntVector var(1,0);
  createTape(var);
}

void ThermalBlockBC1::eval(adoublev& x,
adoublev& vp, adoublev& ip)
{
  // x[0]: (T/T_0)^4 - 1
  vp[0] = T0 * pow(x[0] + one, .25);
  ip[0] = T0*T0*T0*T0 * 0.0002777777 * epsilon*sigma* x[0];
  if (hn)
    ip[0] += hn * pow(vp[0]-T0,1.25);
  if (hf)
    ip[0] += hf * (vp[0]-T0);
}

