// This may look like C code, but it is really -*- C++ -*-
//
// This is a pulsed voltage source with exponential taper 
// (transient analysis only)
//
//       n1     + ---  -   n2
//          o----(   )-----o
//                ---     
// This element behaves as a short circuit for AC/HB analysis.
//
//
// by Frank P. Hart 1/05/2005

#ifndef VpulseAg_h
#define VpulseAg_h 1

class VpulseAg : public Element
{
public:
  
  VpulseAg(const string& iname);

  ~VpulseAg() {}

  static const char* getNetlistName()
  {
    return einfo.name;
  }

  // This element adds equations to the MNAM
  virtual unsigned getExtraRC(const unsigned& eqn_number, 
			      const MNAMType& type);
  virtual void getExtraRC(unsigned& first_eqn, unsigned& n_rows) const;

  // fill MNAM
  virtual void fillMNAM(FreqMNAM* mnam);
  virtual void fillMNAM(TimeMNAM* mnam);
  virtual void fillSourceV(TimeMNAM* mnam);

  // State variable transient analysis
  virtual void svTran(TimeDomainSV *tdsv);
  virtual void deriv_svTran(TimeDomainSV *tdsv);


private:

  // row assigned to this instance by the FreqMNAM
  unsigned my_row;

  // Time for the integer number of periods before current time
  double int_per_time;

  // Element information
  static ItemInfo einfo;

  // Number of parameters of this element
  static const unsigned n_par;

  // Envelope parameter variables
  double v1, v2, td, tr, tf, pw, per;
  int ec;

  // Boolean parameters to null portions of 
  // signal for spectral investigation
  bool null_rise, null_chirp, null_fall;

  // AC signal parameter variables
  double fo, deltaf, phi;
  int chirpdir;

  // Parameter information
  static ParmInfo pinfo[];

  // Boolean covering global initialization 
  // after parameter values available
  // Executes conditionally only once in ::fillsource
  bool do_source_init;

  // Time constants derived from tr and tf
  // Normalized non-zero rail voltage vn.
  double tau_r, tau_f, vn;

  // Pulse width high and low vars for better readability
  double pwl, pwh;

  // Chirp rate (Hz/time)
  double mu;

  // Chirp frequency values at pulse extremities
  double fchrp_tr, fchrp_tf;

  // Constant phase and boundary phase corrections
  double phir, phib_tr, phib_tf;

  // Booleans to governing execution at chirp boundaries
  bool bnd_tr, bnd_tf;

};

#endif


