// AdolcElement is the base class for nonlinear elements which
// use Adol-C. The derived class supplies a suitable eval() routine.
// by Carlos E. Christoffersen

#ifndef AdolcElement_h
#define AdolcElement_h 1

#include "Element.h"
#include "adouble.h"
#include <cassert>

class TimeDomainSV;
class FreqDomainSV;
class WaveletDomainSV;

//-----------------------------------------------------------------------

class AdolcElement : public Element
{
	public:

  AdolcElement(ItemInfo* einfo, ParmInfo* param_desc, const int& numparms,
	const string& iname);

  virtual ~AdolcElement();

  // ---------------------------------------
  // Analysis Methods
  // ---------------------------------------

  // State variable HB.
  virtual void svHB(FreqDomainSV* fdsv);
  virtual void deriv_svHB(FreqDomainSV* fdsv);

  // State variable transient.
  virtual void svTran(TimeDomainSV* tdsv);
  virtual void deriv_svTran(TimeDomainSV* tdsv);

  // State variable wavelet transient
  virtual void svWav(WaveletDomainSV* wdsv);
  virtual void deriv_svWav(WaveletDomainSV* wdsv);

  // Return the element's delay vector
  virtual DenseDoubleVector getDelayVec();

	protected:

  // Select required input variables and create the ADOL-C tape We
  // provide reduced versions for elements that not require
  // derivatives or only require first-order derivatives.
  void createTape(const DenseIntVector& var);

  void createTape(const DenseIntVector& var,
	const DenseIntVector& dvar);

  void createTape(const DenseIntVector& var,
	const DenseIntVector& dvar,
	const DenseIntVector& d2var,
	const DenseIntVector& t_var,
	const DenseDoubleVector& delay);

  // This function is useful for seccionally defined models
  void reTape();

  // return the current time of simulation
  inline double getCurrentTime() { return currentTime; }

  // Generic state variable evaluation routine
  // The meaning of each x component is assigned in
  // createTape()
  virtual void eval(adoublev& x, adoublev& vp, adoublev& ip) = 0;

	private:

  // Number of states
  unsigned my_nstates;
  DenseIntVector var, dvar, d2var, t_var;
  DenseDoubleVector delay;
  // Aux variables to save additions
  int d2const, delconst;
  // Auxiliary arrays for built-in analysis evaluation routines
  static double *x, *f, *jac_col_sel;
  // Default size of allocated memory
  static int aux_size;
  // Flag to signal if the function is taped (ADOL-C tape)
  bool taped;
  int ni_adolc, nd_adolc;
  // This flag signal if memory is already allocated.
  static bool SVWSstatus;
  void allocSVWS();
  // Auxiliary functions for HB methods
  void svHB_timeX(FreqDomainSV*& fdsv);
  void svHB_getx(FreqDomainSV*& fdsv, const int& idx);

  // current time in transient simulation
  double currentTime;
};

#endif

